/**
 *
 * $Id: MainW.c,v 1.39 1998/09/17 10:25:41 kbwj Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static const char rcsid[] = "$Id: MainW.c,v 1.39 1998/09/17 10:25:41 kbwj Exp $";

#include <LTconfig.h>
#include <XmI/XmI.h>
#include <stdio.h>
#include <Xm/XmP.h>
#include <Xm/ScrollBarP.h>
#include <Xm/ScrolledWP.h>
#include <Xm/RowColumnP.h>
#include <Xm/CommandP.h>
#include <Xm/RepType.h>
#include <Xm/MainWP.h>
#include <Xm/SeparatoG.h>
#include <Xm/SeparatoGP.h>
#include <X11/Xfuncs.h>

#include <XmI/DebugUtil.h>

#define	VALID(w)	(w != NULL && XtIsManaged(w))

/* Forward Declarations */

#if 0
static void class_initialize();
#endif

static void class_part_initialize(WidgetClass w_class);

static void initialize(Widget request, Widget new_w,
		       ArgList args, Cardinal *num_args);

static void resize(Widget w);

static XtGeometryResult query_geometry(Widget w,
				       XtWidgetGeometry *proposed,
				       XtWidgetGeometry *answer);

static Boolean set_values(Widget current, Widget request, Widget new_w,
			  ArgList args, Cardinal *num_args);

static XtGeometryResult geometry_manager(Widget w,
					 XtWidgetGeometry *request,
					 XtWidgetGeometry *reply);

static void change_managed(Widget w);

static void insert_child(Widget w);


void _XmRepositionScrolledWindow(Widget w,
				 XtPointer client,
				 XtPointer call);


static void ReparentChild(Widget w, Widget child);

/*
 * Resources for the MainWindow class
 */
#define Offset(field) XtOffsetOf(XmMainWindowRec, mwindow.field)
static XtResource resources[] =
{
    {
	XmNcommandWindow, XmCCommandWindow, XmRWidget,
	sizeof(Widget), Offset(CommandWindow),
	XmRImmediate, NULL
    },
    {
	XmNcommandWindowLocation, XmCCommandWindowLocation,
	XmRCommandWindowLocation,
	sizeof(unsigned char), Offset(CommandLoc),
	XmRImmediate, (XtPointer)XmCOMMAND_ABOVE_WORKSPACE
    },
    {
	XmNmenuBar, XmCMenuBar, XmRWidget,
	sizeof(Widget), Offset(MenuBar),
	XmRImmediate, NULL
    },
    {
	XmNmessageWindow, XmCMessageWindow, XmRWidget,
	sizeof(Widget), Offset(Message),
	XmRImmediate, NULL
    },
    {
	XmNmainWindowMarginWidth, XmCMainWindowMarginWidth,
	XmRHorizontalDimension,
	sizeof(Dimension), Offset(margin_width),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNmainWindowMarginHeight, XmCMainWindowMarginHeight,
	XmRVerticalDimension,
	sizeof(Dimension), Offset(margin_height),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNshowSeparator, XmCShowSeparator, XmRBoolean,
	sizeof(Boolean), Offset(ShowSep),
	XtRImmediate, (XtPointer)False
    }
};

static XmSyntheticResource syn_resources[] =
{
    {
	XmNmainWindowMarginWidth,
	sizeof(Dimension), Offset(margin_width),
	_XmFromHorizontalPixels, _XmToHorizontalPixels
    },
    {
	XmNmainWindowMarginHeight,
	sizeof(Dimension), Offset(margin_height),
	_XmFromVerticalPixels, _XmToVerticalPixels
    }
};

/* Add Actions and Translations -- FIX ME */
/* *INDENT-OFF* */
#if 0
static XmBaseClassExtRec _XmMainWindowCoreClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,                             
    /* version                   */ XmBaseClassExtVersion,
    /* size                      */ sizeof(XmBaseClassExtRec),
    /* initialize_prehook        */ NULL,
    /* set_values_prehook        */ NULL,
    /* initialize_posthook       */ NULL,
    /* set_values_posthook       */ NULL,
    /* secondary_object_class    */ NULL,
    /* secondary_object_create   */ NULL,
    /* get_secondary_resources   */ NULL,
    /* fast_subclass             */ { 0 },
    /* get_values_prehook        */ NULL,
    /* get_values_posthook       */ NULL,
    /* class_part_init_prehook   */ NULL,
    /* class_part_init_posthook  */ NULL,
    /* ext_resources             */ NULL,
    /* compiled_ext_resources    */ NULL,
    /* num_ext_resources         */ 0,
    /* use_sub_resources         */ False,
    /* widget_navigable          */ XmInheritWidgetNavigable,
    /* focus_change              */ XmInheritFocusChange,
    /* wrapper_data              */ NULL
};

static XmManagerClassExtRec _XmMainWMClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,
    /* version                   */ XmManagerClassExtVersion,
    /* record_size               */ sizeof(XmManagerClassExtRec),
    /* traversal_children        */ NULL /* FIX ME */
};
#endif

XmMainWindowClassRec xmMainWindowClassRec = {
    /* Core class part */
    {
	/* superclass            */ (WidgetClass) &xmScrolledWindowClassRec,
        /* class_name            */ "XmMainWindow",
	/* widget_size           */ sizeof(XmMainWindowRec),
	/* class_initialize      */ NULL /*class_initialize*/,
	/* class_part_initialize */ class_part_initialize,
	/* class_inited          */ False,
	/* initialize            */ initialize,
	/* initialize_hook       */ NULL,
	/* realize               */ XtInheritRealize,
	/* actions               */ NULL,
	/* num_actions           */ 0,
	/* resources             */ resources,
	/* num_resources         */ XtNumber(resources),
	/* xrm_class             */ NULLQUARK,
	/* compress_motion       */ True,
	/* compress_exposure     */ XtExposeCompressSeries,
	/* compress_enterleave   */ True,
	/* visible_interest      */ False,
	/* destroy               */ NULL,
	/* resize                */ resize,
	/* expose                */ XtInheritExpose,
	/* set_values            */ set_values,
	/* set_values_hook       */ NULL,
	/* set_values_almost     */ XtInheritSetValuesAlmost,
	/* get_values_hook       */ NULL,
	/* accept_focus          */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* tm_table              */ XtInheritTranslations /*NULL*/,
	/* query_geometry        */ query_geometry,
	/* display_accelerator   */ NULL,
	/* extension             */ (XtPointer)NULL /*&_XmMainWindowCoreClassExtRec*/
    },
    /* Composite class part */
    {
	/* geometry manager */ geometry_manager, 
        /* change_managed   */ change_managed, 
        /* insert_child     */ insert_child,
        /* delete_child     */ XtInheritDeleteChild,
        /* extension        */ NULL,
    },
    /* Constraint class part */
    {
	/* subresources      */ NULL,
        /* subresource_count */ 0,
        /* constraint_size   */ 0,
        /* initialize        */ NULL,
        /* destroy           */ NULL,
        /* set_values        */ NULL,
        /* extension         */ NULL,
    },
    /* XmManager class part */
    {
	/* translations                 */ XtInheritTranslations,
	/* syn_resources                */ syn_resources,
	/* num_syn_resources            */ XtNumber(syn_resources),
	/* syn_constraint_resources     */ NULL,
	/* num_syn_constraint_resources */ 0,
	/* parent_process               */ XmInheritParentProcess,
	/* extension                    */ (XtPointer)NULL /*&_XmMainWMClassExtRec*/
    },
    /* XmScrolledWindow part */
    {
	/* extension */ NULL,
    },
    /* XmMainWindow part */
    {
	/* extension */ NULL,	
    },
};
/* *INDENT-ON* */

WidgetClass xmMainWindowWidgetClass = (WidgetClass)&xmMainWindowClassRec;

#if 0
static void
class_initialize()
{
    _XmMainWindowCoreClassExtRec.record_type = XmQmotif;
}
#endif

static void
class_part_initialize(WidgetClass widget_class)
{
    _XmFastSubclassInit(widget_class, XmMAIN_WINDOW_BIT);
}

static void
initialize(Widget request, Widget new_w,
	   ArgList args, Cardinal *num_args)
{
    if (!XmRepTypeValidValue(XmRepTypeGetId(XmRCommandWindowLocation),
			     MW_CommandLoc(new_w),
			     new_w))
    {
	MW_CommandLoc(new_w) = XmCOMMAND_ABOVE_WORKSPACE;
    }

    /* Override setting of XmScrolledWindow's Margin[Width|Height] resources */
    SW_MarginWidth(new_w) = MW_MarginWidth(new_w);
    SW_MarginHeight(new_w) = MW_MarginHeight(new_w);
    SW_XOffset(new_w) = MW_MarginWidth(new_w);
    SW_YOffset(new_w) = MW_MarginHeight(new_w);

    MW_Sep1(new_w) = (XmSeparatorGadget)XmCreateSeparatorGadget(new_w,
								"Separator1",
								args,
								*num_args);
    MW_Sep2(new_w) = (XmSeparatorGadget)XmCreateSeparatorGadget(new_w,
								"Separator2",
								args,
								*num_args);
    MW_Sep3(new_w) = (XmSeparatorGadget)XmCreateSeparatorGadget(new_w,
								"Separator3",
								args,
								*num_args);
    XtManageChild((Widget)MW_Sep1(new_w));
    XtManageChild((Widget)MW_Sep2(new_w));
    XtManageChild((Widget)MW_Sep3(new_w));
}

/* Try to resize to a size where the children fits.
 * It is done by first find the widest child, or group of childs, and
 * then ajust the height accordingly.
 */
static void
MakeResizeRequest(Widget w)
{

    XtWidgetGeometry request, answer;
    Dimension width, height, new_width, new_height, wa_h;
    Dimension mb_bw = 0, cw_bw = 0, mw_bw = 0;
    Dimension s1_bw = 0, s2_bw = 0, s3_bw = 0;

    if (SW_VisualPolicy(w) == XmCONSTANT)
    {
	width = XtWidth(w) - 2 * MW_MarginWidth(w);
	wa_h = XtHeight(w) - SW_YOffset(w) - SW_MarginHeight(w);
    }
    else
    {
	request.request_mode = 0;
	_XmScrolledWindowPreferredSize(w, &request, &answer);
	width = answer.width - 2 * MW_MarginWidth(w);
	wa_h = answer.height - SW_YOffset(w) - SW_MarginHeight(w);
    }

    if (VALID(MW_MenuBar(w)))
    {
	mb_bw = XtBorderWidth(MW_MenuBar(w));
	request.request_mode = 0;
	if (XtQueryGeometry(MW_MenuBar(w), &request, &answer) != XtGeometryNo
	    && answer.request_mode & CWWidth)
	{
	    if (answer.width + 2 * mb_bw > width)
	    {
		width = answer.width + 2 * mb_bw; 
	    }
	}
	else if (XtWidth(MW_MenuBar(w)) + 2 * mb_bw > width)
	{
	    width = XtWidth(MW_MenuBar(w)) + 2 * mb_bw; 
	}
    }

    if (VALID(MW_MessageWindow(w)))
    {
	mw_bw = XtBorderWidth(MW_MessageWindow(w));

	request.request_mode = 0;
	if (XtQueryGeometry(MW_MessageWindow(w), &request, &answer) != 
	    XtGeometryNo  && answer.request_mode & CWWidth)
	{
	    if (answer.width + 2 * mw_bw > width)
	    {
		width = answer.width + 2 * mw_bw; 
	    }
	}
	else if (XtWidth(MW_MessageWindow(w)) + 2 * mw_bw > width)
	{
	    width = XtWidth(MW_MessageWindow(w)) + 2 * mw_bw; 
	}
    }

    if (VALID(MW_CommandWindow(w)))
    {
	cw_bw = XtBorderWidth(MW_CommandWindow(w));

	request.request_mode = 0;
	if (XtQueryGeometry(MW_CommandWindow(w), &request, &answer) != 
	    XtGeometryNo  && answer.request_mode & CWWidth)
	{
	    if (answer.width + 2 * cw_bw > width)
	    {
		width = answer.width + 2 * cw_bw; 
	    }
	}
	else if (XtWidth(MW_CommandWindow(w)) + 2 * cw_bw > width)
	{
	    width = XtWidth(MW_CommandWindow(w)) + 2 * cw_bw; 
	}
    }

    height = wa_h;

    if (VALID(MW_MenuBar(w)))
    {
	request.request_mode = CWWidth;
	request.width = width - 2 * mb_bw;
	if (XtQueryGeometry(MW_MenuBar(w), &request, &answer) != XtGeometryNo
	    && answer.request_mode & CWHeight)
	{
	    height += answer.height + 2 * mb_bw;
	}
	else
	{
	    height += XtHeight(MW_MenuBar(w)) + 2 * mb_bw;
	}
    }

    if (VALID(MW_Sep1(w)))
    {
	s1_bw = XtBorderWidth(MW_Sep1(w));

	if (MW_ShowSep(w) && VALID(MW_MenuBar(w)))
	{
	    height += XtHeight(MW_Sep1(w)) + 2 * s1_bw;
	}
    }

    if (VALID(MW_MessageWindow(w)))
    {
	mw_bw = XtBorderWidth(MW_MessageWindow(w));

	request.request_mode = CWWidth;
	request.width = width - 2 * mw_bw;
	if (XtQueryGeometry(MW_MessageWindow(w), &request, &answer) != 
	    XtGeometryNo  && answer.request_mode & CWHeight)
	{
	    height += answer.height + 2 * mw_bw;
	}
	else
	{
	    height += XtHeight(MW_MessageWindow(w)) + 2 * mw_bw;
	}
    }
    if (VALID(MW_Sep3(w)))
    {
	s3_bw = XtBorderWidth(MW_Sep3(w));

	if (MW_ShowSep(w) && VALID(MW_MessageWindow(w)))
	{
	    height += XtHeight(MW_Sep3(w)) + 2 * s3_bw;
	}
    }

    if (VALID(MW_CommandWindow(w)))
    {
	cw_bw = XtBorderWidth(MW_CommandWindow(w));

	request.request_mode = CWWidth;
	request.width = width - 2 * cw_bw;
	if (XtQueryGeometry(MW_CommandWindow(w), &request, &answer) != 
	    XtGeometryNo  && answer.request_mode & CWHeight)
	{
	    height += answer.height + 2 * cw_bw;
	}
	else
	{
	    height += XtHeight(MW_CommandWindow(w)) + 2 * cw_bw;
	}
    }

    if (VALID(MW_Sep2(w)))
    {
	s2_bw = XtBorderWidth(MW_Sep2(w));

	if (MW_ShowSep(w) && VALID(MW_CommandWindow(w)))
	{
	    if (MW_CommandLoc(w) == XmCOMMAND_ABOVE_WORKSPACE)
	    {
		height += XtHeight(MW_Sep2(w)) + 2 * s2_bw;
	    }
	}
    }

    width += 2 * MW_MarginWidth(w);
    height += 2 * MW_MarginHeight(w);

    if (!XtIsRealized(w))
    {
	if (SW_GivenWidth(w) != 0)
	{
	    width = SW_GivenWidth(w);
	}
	if (SW_GivenHeight(w) != 0)
	{
	    height = SW_GivenHeight(w);
	}
    }

    if (XtMakeResizeRequest(w, width, height, &new_width, &new_height) ==
	XtGeometryAlmost)
    {
	XtMakeResizeRequest(w, new_width, new_height, &new_width, &new_height);
    }
}

static Boolean
set_values(Widget old, Widget request, Widget new_w,
	   ArgList args, Cardinal *num_args)
{
    Boolean refresh = False;
#define NE(x)	(x(old) != x(new_w))

    if (NE(MW_CommandLoc))
    {
	if (!XmRepTypeValidValue(XmRepTypeGetId(XmRCommandWindowLocation),
				 MW_CommandLoc(new_w),
				 new_w))
	{
	    MW_CommandLoc(new_w) = MW_CommandLoc(old);
	}
	else
	{
	    refresh = True;
	}
    }

    if (NE(SW_WorkWindow))
    {
        DEBUGOUT(XdbDebug(__FILE__, new_w,
			  "SetValues: Changed the work window !\n"));

	if (MW_CommandWindow(new_w) == SW_WorkWindow(new_w))
	{
	    MW_CommandWindow(new_w) = NULL;
	}
	if (MW_MessageWindow(new_w) == SW_WorkWindow(new_w))
	{
	    MW_MessageWindow(new_w) = NULL;
	}
	if (MW_MenuBar(new_w) == SW_WorkWindow(new_w))
	{
	    MW_MenuBar(new_w) = NULL;
	}

	refresh = True;
    }

    if (NE(MW_CommandWindow))
    {
        DEBUGOUT(XdbDebug(__FILE__, new_w,
			  "SetValues: Changed the command window !\n"));

	if (SW_VisualPolicy(new_w) == XmCONSTANT &&
	    XtParent(MW_CommandWindow(new_w)) == (Widget)SW_ClipWindow(new_w))
	{
	    ReparentChild(new_w, MW_CommandWindow(new_w));
	}

	if (SW_WorkWindow(new_w) == MW_CommandWindow(new_w))
	{
	    SW_WorkWindow(new_w) = NULL;
	}
	if (MW_MessageWindow(new_w) == MW_CommandWindow(new_w))
	{
	    MW_MessageWindow(new_w) = NULL;
	}
	if (MW_MenuBar(new_w) == MW_CommandWindow(new_w))
	{
	    MW_MenuBar(new_w) = NULL;
	}

	refresh = True;
    }

    if (NE(MW_MessageWindow))
    {
        DEBUGOUT(XdbDebug(__FILE__, new_w,
			  "SetValues: Changed the message window !\n"));

	if (SW_VisualPolicy(new_w) == XmCONSTANT &&
	    XtParent(MW_MessageWindow(new_w)) == (Widget)SW_ClipWindow(new_w))
	{
	    ReparentChild(new_w, MW_MessageWindow(new_w));
	}

	if (MW_CommandWindow(new_w) == MW_MessageWindow(new_w))
	{
	    MW_CommandWindow(new_w) = NULL;
	}
	if (SW_WorkWindow(new_w) == MW_MessageWindow(new_w))
	{
	    SW_WorkWindow(new_w) = NULL;
	}
	if (MW_MenuBar(new_w) == MW_MessageWindow(new_w))
	{
	    MW_MenuBar(new_w) = NULL;
	}

	refresh = True;
    }

    if (NE(MW_MenuBar))
    {
        DEBUGOUT(XdbDebug(__FILE__, new_w,
			  "SetValues: Changed the menu bar !\n"));

	if (SW_VisualPolicy(new_w) == XmCONSTANT &&
	    XtParent(MW_MenuBar(new_w)) == (Widget)SW_ClipWindow(new_w))
	{
	    ReparentChild(new_w, MW_MenuBar(new_w));
	}

	if (MW_CommandWindow(new_w) == MW_MenuBar(new_w))
	{
	    MW_CommandWindow(new_w) = NULL;
	}
	if (SW_WorkWindow(new_w) == MW_MenuBar(new_w))
	{
	    SW_WorkWindow(new_w) = NULL;
	}
	if (MW_MessageWindow(new_w) == MW_MenuBar(new_w))
	{
	    MW_MessageWindow(new_w) = NULL;
	}

	refresh = True;
    }

    if (NE(MW_ShowSep) ||
	NE(MW_MarginHeight) ||
	NE(MW_MarginWidth))
    {
	refresh = True;
    }
    if (refresh)
    {
	(*XtClass(new_w)->core_class.resize) (new_w);
    }

    return refresh;
}

/* The layout is made both from top and bottom. All widgets above and
 * inclusive workArea is laid out from the top and the rest from the
 * bottom.
 */

static void
Layout(Widget w)
{

    XtWidgetGeometry request, answer;
    Dimension usable_width, usable_height;
    int cury_from_top, cury_from_bottom, curx;
    XmMWValues vals;
    Dimension mb_bw = 0, cw_bw = 0, mw_bw = 0;
    Dimension s1_bw = 0, s2_bw = 0, s3_bw = 0;

    usable_width = XtWidth(w) - 2 * MW_MarginWidth(w);
    usable_height = XtHeight(w) - 2 * MW_MarginHeight(w);


    curx = MW_MarginWidth(w);
    cury_from_top = MW_MarginHeight(w);
    cury_from_bottom = XtHeight(w) - MW_MarginHeight(w);


    if (VALID(MW_MenuBar(w)))
    {
	mb_bw = XtBorderWidth(MW_MenuBar(w));
	request.request_mode = CWWidth;
	request.width = usable_width - 2 * mb_bw;
	if (XtQueryGeometry(MW_MenuBar(w), &request, &answer) != XtGeometryNo
	    && answer.request_mode & CWHeight)
	{
	    vals.mbh = answer.height;
	}
	else
	{
	    vals.mbh = XtHeight(MW_MenuBar(w));
	}
	vals.mbw = usable_width - 2 * mb_bw;
	vals.mbx = curx;
	vals.mby = cury_from_top;
	cury_from_top += vals.mbh + 2 * mb_bw;
    }

    if (VALID(MW_Sep1(w)))
    {
	s1_bw = XtBorderWidth(MW_Sep1(w));

	if (MW_ShowSep(w) && VALID(MW_MenuBar(w)))
	{
	    vals.s1x = curx;
	    vals.s1y = cury_from_top;
	    vals.s1w = usable_width - 2 * s1_bw;
	    vals.s1h = XtHeight(MW_Sep1(w));
	    cury_from_top += vals.s1h + 2 * s1_bw;
	}
	else
	{
	    vals.s1x = XtWidth(w);
	    vals.s1y = cury_from_top;
	    vals.s1w = usable_width - 2 * s1_bw;
	    vals.s1h = XtHeight(MW_Sep1(w));
	}
    }

    if (VALID(MW_MessageWindow(w)))
    {
	mw_bw = XtBorderWidth(MW_MessageWindow(w));

	request.request_mode = CWWidth;
	request.width = usable_width - 2 * mw_bw;
	if (XtQueryGeometry(MW_MessageWindow(w), &request, &answer) != 
	    XtGeometryNo  && answer.request_mode & CWHeight)
	{
	    vals.mwh = answer.height;
	}
	else
	{
	    vals.mwh = XtHeight(MW_MessageWindow(w));
	}
	vals.mww = usable_width - 2 * mw_bw;
	vals.mwx = curx;
	vals.mwy = cury_from_bottom - vals.mwh - 2 * mw_bw;
	cury_from_bottom -= vals.mwh + 2 * mw_bw;
    }
    if (VALID(MW_Sep3(w)))
    {
	s3_bw = XtBorderWidth(MW_Sep3(w));

	if (MW_ShowSep(w) && VALID(MW_MessageWindow(w)))
	{
	    vals.s3w = usable_width - 2 * s3_bw;
	    vals.s3h = XtHeight(MW_Sep3(w));
	    vals.s3x = curx;
	    vals.s3y = cury_from_bottom - vals.s3h - 2 * s3_bw;
	    cury_from_bottom -= vals.s3h + 2 * s3_bw;
	}
	else
	{
	    vals.s3w = usable_width - 2 * s3_bw;
	    vals.s3h = XtHeight(MW_Sep3(w));
	    vals.s3x = XtWidth(w);
	    vals.s3y = cury_from_bottom - vals.s3h - 2 * s3_bw;
	}
    }

    if (VALID(MW_CommandWindow(w)))
    {
	cw_bw = XtBorderWidth(MW_CommandWindow(w));

	request.request_mode = CWWidth;
	request.width = usable_width - 2 * cw_bw;
	if (XtQueryGeometry(MW_CommandWindow(w), &request, &answer) != 
	    XtGeometryNo  && answer.request_mode & CWHeight)
	{
	    vals.cwh = answer.height;
	}
	else
	{
	    vals.cwh = XtHeight(MW_CommandWindow(w));
	}
	vals.cww = usable_width - 2 * cw_bw;
	vals.cwx = curx;
	if (MW_CommandLoc(w) == XmCOMMAND_ABOVE_WORKSPACE)
	{
	    vals.cwy = cury_from_top;
	    cury_from_top += vals.cwh + 2 * cw_bw;
	}
	else
	{
	    vals.cwy = cury_from_bottom - vals.cwh - 2 * cw_bw;
	    cury_from_bottom -= vals.cwh + 2 * cw_bw;
	}
    }

    if (VALID(MW_Sep2(w)))
    {
	s2_bw = XtBorderWidth(MW_Sep2(w));

	if (MW_ShowSep(w) && VALID(MW_CommandWindow(w)))
	{
	    if (MW_CommandLoc(w) == XmCOMMAND_ABOVE_WORKSPACE)
	    {
		vals.s2w = usable_width - 2 * s2_bw;
		vals.s2h = XtHeight(MW_Sep2(w));
		vals.s2x = curx;
		vals.s2y = cury_from_top;
		cury_from_top += vals.s2h + 2 * s2_bw;
	    }
	    else
	    {
		vals.s2w = usable_width - 2 * s2_bw;
		vals.s2h = XtHeight(MW_Sep2(w));
		vals.s2x = curx;
		vals.s2y = cury_from_bottom - vals.s2h - 2 * s2_bw;
		cury_from_bottom -= vals.s2h + 2 * s2_bw;
	    }
	}
	else
	{
	    vals.s2w = usable_width - 2 * s2_bw;
	    vals.s2h = XtHeight(MW_Sep2(w));
	    vals.s2x = XtWidth(w);
	    vals.s2y = cury_from_top;
	}
    }

    if (VALID(MW_MenuBar(w)))
    {
	_XmConfigureObject(MW_MenuBar(w), vals.mbx, vals.mby,
			   vals.mbw, vals.mbh, mb_bw);
    }
    if (VALID(MW_Sep1(w)))
    {
	_XmConfigureObject((Widget) MW_Sep1(w), vals.s1x, vals.s1y,
			   vals.s1w, vals.s1h, s1_bw);
    }
    if (VALID(MW_CommandWindow(w)))
    {
	_XmConfigureObject(MW_CommandWindow(w), vals.cwx, vals.cwy,
			   vals.cww, vals.cwh, cw_bw);
    }
    if (VALID(MW_Sep2(w)))
    {
	_XmConfigureObject((Widget) MW_Sep2(w), vals.s2x, vals.s2y,
			   vals.s2w, vals.s2h, s2_bw);
    }
    if (VALID(MW_MessageWindow(w)))
    {
	_XmConfigureObject(MW_MessageWindow(w), vals.mwx, vals.mwy,
			   vals.mww, vals.mwh, mw_bw);
    }
    if (VALID(MW_Sep3(w)))
    {
	_XmConfigureObject((Widget) MW_Sep3(w), vals.s3x, vals.s3y,
			   vals.s3w, vals.s3h, s3_bw);
    }
    SW_XOffset(w) = MW_MarginWidth(w);
    SW_MarginWidth(w) = MW_MarginWidth(w);
    SW_YOffset(w) = cury_from_top;
    SW_MarginHeight(w) = XtHeight(w) - cury_from_bottom;

    (*xmScrolledWindowClassRec.core_class.resize) (w);
}

static void
resize(Widget w)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "Resize (w %d h %d)\n",
		      XtWidth(w), XtHeight(w)));

    SW_FromResize(w) = True;

    Layout(w);

    SW_FromResize(w) = False;
}


static XtGeometryResult
query_geometry(Widget w,
	       XtWidgetGeometry *intended,
	       XtWidgetGeometry *preferred)
{
    XtWidgetGeometry wants;
    XmMWValues vals;

    wants = *intended;

    /* _XmMainWindowPreferredSize(w, w, NULL, &vals); */

    if (preferred)
    {
	preferred->width = vals.MwW;
	preferred->height = vals.MwH;
    }

    if (((wants.request_mode & CWWidth) &&
	 wants.width == preferred->width) &&
	((wants.request_mode & CWHeight) &&
	 wants.height == preferred->height))
    {
	return XtGeometryNo;
    }

    if ((wants.request_mode & CWWidth) &&
	wants.width != preferred->width)
    {
	return XtGeometryAlmost;	/* Something's different */
    }
    if ((wants.request_mode & CWHeight) &&
	wants.height != preferred->height)
    {
	return XtGeometryAlmost;	/* Something's different */
    }

    return XtGeometryYes;
}

static void
CreateManagedList(Widget w, Widget **managed, int *num_managed,
		  int *num_allocated)
{
    if (XtIsComposite(w) && MGR_NumChildren(w) > 0)
    {
	int i;
	for (i = 0; i < MGR_NumChildren(w); i++)
	{
	    CreateManagedList(MGR_Children(w)[i], managed, num_managed,
			      num_allocated);
	}
    }
    if (XtIsManaged(w))
    {
	if (*num_managed == *num_allocated)
	{
	/* Allocate more space */
	*num_allocated += (*num_allocated / 2) + 2;
	*managed =  
	    (WidgetList) XtRealloc((XtPointer) *managed,
				   (unsigned) (*num_allocated *
					       sizeof(Widget)));
	}
	*managed[*num_managed++] = w;
    }
}

static void
ReparentChild(Widget w, Widget child)
{

    Widget old_parent = XtParent(child), *managed;
    int i, j, num_managed, num_allocated;
    Boolean is_realized;

    if (old_parent == w)
    {
	return;
    }
    /* first make a list of children who needs to be managed again */
    num_managed = 0;
    num_allocated = 2;
    managed = (Widget *) XtMalloc(num_allocated * sizeof(Widget));
    CreateManagedList(child, &managed, & num_managed, &num_allocated);
    for (i = 0; i < MGR_NumChildren(old_parent); i++)
    {
	if (MGR_Children(old_parent)[i] == child)
	{
	    break;
	}
    }
    if (MGR_NumChildren(w) == MGR_NumSlots(w)) {
	/* Allocate more space */
	MGR_NumSlots(w) += (MGR_NumSlots(w) / 2) + 2;
	MGR_Children(w) =  
	    (WidgetList) XtRealloc((XtPointer) MGR_Children(w),
				   (unsigned) (MGR_NumSlots(w) *
					       sizeof(Widget)));
    }
    if ((is_realized = XtIsRealized(child)))
    {
	XtUnrealizeWidget(child);
    } 
    MGR_Children(w)[MGR_NumChildren(w)++] = child;
    XtParent(child) = w;
    if (is_realized)
    {
	XtRealizeWidget(child);
    }
    for (j = 0; j < num_managed; j++)
    {
	XtManageChild(managed[j]);
    }
    XtFree((char *) managed);
 
    for (j = i + 1; j < MGR_NumChildren(old_parent); j++)
    {
	MGR_Children(old_parent)[j - 1] = MGR_Children(old_parent)[j];
    }
    MGR_NumChildren(old_parent)--;
}




static void
change_managed(Widget w)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "ChangeManaged\n"));
    /*
     * It is possible that some of the MainWindows children is made children
     * of ClipWindow by the ScrolledWindow. Steal them back.
     */
    if (SW_VisualPolicy(w) == XmCONSTANT &&
	MGR_NumChildren(SW_ClipWindow(w)) > 1)
    {
	XmDrawingAreaWidget cw = SW_ClipWindow(w);
	Widget mw_children[3];
	int i, num_mw_children = 0;

	for (i = 0 ; i < MGR_NumChildren(cw); i++)
	{
	    if (MGR_Children(cw)[i] == MW_CommandWindow(w) ||
		MGR_Children(cw)[i] == MW_MenuBar(w) ||
		MGR_Children(cw)[i] == MW_MessageWindow(w))
	    {
		mw_children[num_mw_children++] = MGR_Children(cw)[i];
	    }
	}
	for (i = 0; i < num_mw_children; i++)
	{
	    ReparentChild(w, mw_children[i]);
	}
    }
    MakeResizeRequest(w);
    Layout(w);
}


static XtGeometryResult
geometry_manager(Widget w,
		 XtWidgetGeometry *desired,
		 XtWidgetGeometry *allowed)
{
    XtWidgetGeometry wants;

    DEBUGOUT(XdbDebug2(__FILE__, XtParent(w), w,
		       "geometry_manager request %s\n",
		       XdbWidgetGeometry2String(desired)));

#define Wants(flag)     (wants.request_mode & flag)

    /* If it isnt one of the childs i know, hand it over to ScrolledWindow */

    if (w != (Widget) MW_MenuBar(w) &&
	w != MW_CommandWindow(w) &&
	w != MW_MessageWindow(w) &&
	w != (Widget) MW_Sep1(w) &&
	w != (Widget) MW_Sep2(w) &&
	w != (Widget) MW_Sep3(w))
    {
	return (*xmScrolledWindowClassRec.composite_class.geometry_manager)
	    (w, desired, allowed);
    }

    wants = *desired;
    *allowed = *desired;
    return XtGeometryYes;
}

Widget
XmCreateMainWindow(Widget parent,
		   char *name,
		   Arg *argList,
		   Cardinal argcount)
{
    return XtCreateWidget(name,
			  xmMainWindowWidgetClass,
			  parent,
			  argList, argcount);

}

Widget
XmMainWindowSep1(Widget widget)
{
    return (Widget)MW_Sep1(widget);
}

Widget
XmMainWindowSep2(Widget widget)
{
    return (Widget)MW_Sep2(widget);
}

Widget
XmMainWindowSep3(Widget widget)
{
    return (Widget)MW_Sep3(widget);
}

void
XmMainWindowSetAreas(Widget widget,
		     Widget menu_bar,
		     Widget command_window,
		     Widget horizontal_scrollbar,
		     Widget vertical_scrollbar,
		     Widget work_region)
{
    Arg args[5];
    int n = 0;

    if (menu_bar)
    {
	XtSetArg(args[n], XmNmenuBar, MW_MenuBar(widget)); n++;
    }
    if (command_window)
    {
	XtSetArg(args[n], XmNcommandWindow, MW_CommandWindow(widget)); n++;
    }
    if (work_region)
    {
	XtSetArg(args[n], XmNworkWindow, SW_WorkWindow(widget)); n++;
    }
    if (horizontal_scrollbar)
    {
	XtSetArg(args[n], XmNhorizontalScrollBar, SW_HSB(widget)); n++;
    }
    if (vertical_scrollbar)
    {
	XtSetArg(args[n], XmNverticalScrollBar, SW_VSB(widget)); n++;
    }
    if (n > 0)
    {
	XtSetValues(widget, args, n);
    }

}

static void
insert_child(Widget w)
{
    Widget p = XtParent(w);

    /* If we is in initialize then asume initial widgets and let ScrolledW
     * handle this
     */
    if (SW_InInit(p))
    { 
	(*xmScrolledWindowClassRec.composite_class.insert_child) (w);
	return;
    }


    if ((XmIsRowColumn(w) && RC_Type(w) == XmMENU_BAR) ||
	XmIsSeparator(w) || XmIsSeparatorGadget(w))
    {
	(*xmManagerClassRec.composite_class.insert_child) (w);

	if (XmIsRowColumn(w) && RC_Type(w) == XmMENU_BAR)
	{
	    DEBUGOUT(XdbDebug2(__FILE__, p, w,
			       "insert_child : this is the menu bar\n"));
	    MW_MenuBar(p) = w;
	}
	else if (XmIsSeparator(w) || XmIsSeparatorGadget(w))
	{
	    DEBUGOUT(XdbDebug2(__FILE__, p, w,
			       "insert_child : this is a separator\n"));
	}
	else if (MW_MessageWindow(p) == NULL &&
	    w != MW_CommandWindow(p) &&
	    w != MW_MenuBar(p) &&
	    w != SW_WorkWindow(p) &&
	    w != (Widget)SW_ClipWindow(p) &&
	    w != (Widget)SW_HSB(p) &&
	    w != (Widget)SW_VSB(p))
	{
	    DEBUGOUT(XdbDebug2(__FILE__, p, w,
			       "insert_child : this is the message window\n"));
	    MW_MessageWindow(p) = w;
	}
    }
    else
    {
	/* 
	 * Now we have a scrollbar or a workwindow - let scrolledwindow
	 * handle this
	 */
	(*xmScrolledWindowClassRec.composite_class.insert_child) (w);

	if (MW_MessageWindow(p) == NULL &&
	    w != MW_CommandWindow(p) &&
	    w != MW_MenuBar(p) &&
	    w != SW_WorkWindow(p) &&
	    w != (Widget)SW_ClipWindow(p) &&
	    w != (Widget)SW_HSB(p) &&
	    w != (Widget)SW_VSB(p))
	{
	    DEBUGOUT(XdbDebug2(__FILE__, p, w,
			       "insert_child : this is the message window\n"));
	    MW_MessageWindow(p) = w;
	}
    }
}

