/*
 *
 * $Id: ScrolledW.c,v 1.89 1998/09/18 18:24:16 pgw Exp $
 *
 * Copyright (C) 1996 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static const char rcsid[] = "$Id: ScrolledW.c,v 1.89 1998/09/18 18:24:16 pgw Exp $";

/*
 * Current status of changes in layout (Danny 18/4/1997) :
 *
 * the bottom_right (default) and bottom_left cases are believed to be finished.
 * the top_left and top_right cases are badly broken
 */

#include <LTconfig.h>
#include <XmI/XmI.h>

#include <stdio.h>

#include <X11/Xfuncs.h>
#include <Xm/XmP.h>
#include <Xm/ScrolledWP.h>
#include <Xm/ScrollBarP.h>
#include <Xm/DrawingAP.h>
#include <Xm/RowColumnP.h>
#include <Xm/TransltnsP.h>
#include <Xm/RepType.h>
#include <Xm/VirtKeysP.h>

#include <XmI/DebugUtil.h>
#include <XmI/MessagesI.h>

#ifndef XmUNSPECIFIED
#define XmUNSPECIFIED (~0)
#endif


/* Forward Declarations */

#if 1
static void class_initialize();
#endif

static void class_part_initialize(WidgetClass w_class);

static void initialize(Widget request, Widget new_w,
		       ArgList args, Cardinal *num_args);

static void resize(Widget w);

static void expose(Widget w, XEvent *event, Region region);

static XtGeometryResult query_geometry(Widget w,
				       XtWidgetGeometry *proposed,
				       XtWidgetGeometry *answer);

static Boolean set_values(Widget current, Widget request, Widget new_w,
			  ArgList args, Cardinal *num_args);

static void change_managed(Widget w);

static void insert_child(Widget w);

static void delete_child(Widget w);

static XtGeometryResult geometry_manager(Widget w,
					 XtWidgetGeometry *request,
					 XtWidgetGeometry *reply);


static void HsbCallback(Widget w, XtPointer client, XtPointer call);

static void VsbCallback(Widget w, XtPointer client, XtPointer call);

#if XmVERSION > 1
void _XmInitializeScrollBars(Widget w);
#endif

/*
 * Resources for the ScrolledWindow class
 */
#define Offset(field) XtOffsetOf(XmScrolledWindowRec, swindow.field)
#define MGR_Offset(field) XtOffsetOf(XmScrolledWindowRec, manager.field)
static XtResource resources[] =
{
    {
	XmNhorizontalScrollBar, XmCHorizontalScrollBar, XmRWidget,
	sizeof(Widget), Offset(hScrollBar),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNverticalScrollBar, XmCVerticalScrollBar, XmRWidget,
	sizeof(Widget), Offset(vScrollBar),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNworkWindow, XmCWorkWindow, XmRWidget,
	sizeof(Widget), Offset(WorkWindow),
	XmRImmediate, (XtPointer)NULL
    },
    {
	XmNclipWindow, XmCClipWindow, XmRWidget,
	sizeof(Widget), Offset(ClipWindow),
	XtRImmediate, (XtPointer)NULL
    },
    {
	XmNscrollingPolicy, XmCScrollingPolicy, XmRScrollingPolicy,
	sizeof(unsigned char), Offset(ScrollPolicy),
	XmRImmediate, (XtPointer)XmAPPLICATION_DEFINED
    },
    {
	XmNvisualPolicy, XmCVisualPolicy, XmRVisualPolicy,
	sizeof(unsigned char), Offset(VisualPolicy),
	XmRImmediate, (XtPointer)XmVARIABLE
    },
    {
	XmNscrollBarDisplayPolicy, XmCScrollBarDisplayPolicy,
	XmRScrollBarDisplayPolicy,
	sizeof(unsigned char), Offset(ScrollBarPolicy),
	XmRImmediate, (XtPointer)((unsigned char)XmUNSPECIFIED)
    },
    {
	XmNscrollBarPlacement, XmCScrollBarPlacement, XmRScrollBarPlacement,
	sizeof(unsigned char), Offset(Placement),
	XtRImmediate, (XtPointer)XmBOTTOM_RIGHT
    },
    {
	XmNscrolledWindowMarginWidth, XmCScrolledWindowMarginWidth,
	XmRHorizontalDimension,
	sizeof(Dimension), Offset(WidthPad),
	XtRImmediate, (XtPointer)0
    },
    {
	XmNscrolledWindowMarginHeight, XmCScrolledWindowMarginHeight,
	XmRVerticalDimension,
	sizeof(Dimension), Offset(HeightPad),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNspacing, XmCSpacing, XmRHorizontalDimension,
	sizeof(Dimension), Offset(pad),
	XtRImmediate, (XtPointer)XmINVALID_DIMENSION
    },
    {
	XmNshadowThickness, XmCShadowThickness, XmRHorizontalDimension,
	sizeof(Dimension), MGR_Offset(shadow_thickness),
	XmRImmediate, (XtPointer)XmINVALID_DIMENSION
    },
    {
	XmNtraverseObscuredCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(traverseObscuredCallback),
	XmRImmediate, NULL
    }
};

static XmSyntheticResource syn_resources[] =
{
    {
	XmNscrolledWindowMarginWidth,
	sizeof(Dimension), Offset(WidthPad),
	_XmFromHorizontalPixels, _XmToHorizontalPixels
    },
    {
	XmNscrolledWindowMarginHeight,
	sizeof(Dimension), Offset(HeightPad),
	_XmFromVerticalPixels, _XmToVerticalPixels
    },
    {
	XmNspacing,
	sizeof(Dimension), Offset(pad),
	_XmFromVerticalPixels, NULL
    }
};

static void PageUp(Widget w, XEvent *event,
		   String *params, Cardinal *num_params);

static void PageDown(Widget w, XEvent *event,
		     String *params, Cardinal *num_params);

static void PageLeft(Widget w, XEvent *event,
		     String *params, Cardinal *num_params);

static void PageRight(Widget w, XEvent *event,
		      String *params, Cardinal *num_params);

static void BeginLine(Widget w, XEvent *event,
		      String *params, Cardinal *num_params);

static void EndLine(Widget w, XEvent *event,
		    String *params, Cardinal *num_params);

static void BeginData(Widget w, XEvent *event,
		      String *params, Cardinal *num_params);

static void EndData(Widget w, XEvent *event,
		    String *params, Cardinal *num_params);

static void PageUpGrab(Widget w, XEvent *event,
		       String *params, Cardinal *num_params);

static void PageDownGrab(Widget w, XEvent *event,
		         String *params, Cardinal *num_params);

static void PageLeftGrab(Widget w, XEvent *event,
		         String *params, Cardinal *num_params);

static void PageRightGrab(Widget w, XEvent *event,
		          String *params, Cardinal *num_params);

static void BeginLineGrab(Widget w, XEvent *event,
		          String *params, Cardinal *num_params);

static void EndLineGrab(Widget w, XEvent *event,
		        String *params, Cardinal *num_params);

static void BeginDataGrab(Widget w, XEvent *event,
		          String *params, Cardinal *num_params);

static void EndDataGrab(Widget w, XEvent *event,
		        String *params, Cardinal *num_params);

static void SWNoop(Widget w, XEvent *event,
		   String *params, Cardinal *num_params);
static void CWMapNotify(Widget w, XEvent *event,
		   String *params, Cardinal *num_params);

typedef struct _TransLtnsRec
{
    char*      virtkey;
    unsigned long   modifiers;
    char* function;
} *XmTranslations, XmTranslationRec;

static XmTranslationRec ClipWindowTranslations[] =
{
    {"<Key>osfPageDown",	ControlMask,	"SWRightPageGrab()"},
    {"<Key>osfPageDown",	0,		"SWDownPageGrab()"},
    {"<Key>osfBeginLine",	ControlMask,	"SWTopLineGrab()"},
    {"<Key>osfBeginLine",	0,		"SWBeginLineGrab()"},
    {"<Key>osfEndLine",		ControlMask,	"SWBottomLineGrab()"},
    {"<Key>osfEndLine",		0,		"SWEndLineGrab()"},
    {"<Key>osfPageUp",		ControlMask,	"SWLeftPageGrab()"},
    {"<Key>osfPageUp",		0,		"SWUpPageGrab()"},
    {"<Key>osfHelp",		0,		"Help()"},
    {NULL,			0,		NULL}
};

static XtTranslations RealTranslationsParsed = NULL, TranslationsParsed = NULL;

static XtActionsRec actions[] =
{
    {"SWBeginLine", BeginLine},
    {"SWEndLine", EndLine},
    {"SWTopLine", BeginData},
    {"SWBottomLine", EndData},
    {"SWLeftPage", PageLeft},
    {"SWRightPage", PageRight},
    {"SWUpPage", PageUp},
    {"SWDownPage", PageDown},
    {"SWBeginLineGrab", BeginLineGrab},
    {"SWEndLineGrab", EndLineGrab},
    {"SWTopLineGrab", BeginDataGrab},
    {"SWBottomLineGrab", EndDataGrab},
    {"SWLeftPageGrab", PageLeftGrab},
    {"SWRightPageGrab", PageRightGrab},
    {"SWUpPageGrab", PageUpGrab},
    {"SWDownPageGrab", PageDownGrab},
    {"SWCWMapNotify", CWMapNotify},
    {"SWNoop", SWNoop}
};

/* *INDENT-OFF* */
#if 1
static XmBaseClassExtRec _XmScrolledWCoreClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,                             
    /* version                   */ XmBaseClassExtVersion,
    /* size                      */ sizeof(XmBaseClassExtRec),
    /* initialize_prehook        */ NULL,
    /* set_values_prehook        */ NULL,
    /* initialize_posthook       */ NULL,
    /* set_values_posthook       */ NULL,
    /* secondary_object_class    */ NULL,
    /* secondary_object_create   */ NULL,
    /* get_secondary_resources   */ NULL,
    /* fast_subclass             */ { 0 },
    /* get_values_prehook        */ NULL,
    /* get_values_posthook       */ NULL,
    /* class_part_init_prehook   */ NULL,
    /* class_part_init_posthook  */ NULL,
    /* ext_resources             */ NULL,
    /* compiled_ext_resources    */ NULL,
    /* num_ext_resources         */ 0,
    /* use_sub_resources         */ False,
    /* widget_navigable          */ XmInheritWidgetNavigable,
    /* focus_change              */ XmInheritFocusChange,
    /* wrapper_data              */ NULL
};

static XmManagerClassExtRec _XmScrolledWMClassExtRec = {
    /* next_extension            */ NULL,
    /* record_type               */ NULLQUARK,
    /* version                   */ XmManagerClassExtVersion,
    /* record_size               */ sizeof(XmManagerClassExtRec),
    /* traversal_children        */ NULL /* FIX ME */
};
#endif

XmScrolledWindowClassRec xmScrolledWindowClassRec = {
    /* Core class part */
    {
	/* superclass            */ (WidgetClass) &xmManagerClassRec,
        /* class_name            */ "XmScrolledWindow",
	/* widget_size           */ sizeof(XmScrolledWindowRec),
	/* class_initialize      */ class_initialize,
	/* class_part_initialize */ class_part_initialize,
	/* class_inited          */ False,
	/* initialize            */ initialize,
	/* initialize_hook       */ NULL,
	/* realize               */ XtInheritRealize,
	/* actions               */ actions,
	/* num_actions           */ XtNumber(actions),
	/* resources             */ resources,
	/* num_resources         */ XtNumber(resources),
	/* xrm_class             */ NULLQUARK,
	/* compress_motion       */ True,
	/* compress_exposure     */ XtExposeCompressMaximal,
	/* compress_enterleave   */ True,
	/* visible_interest      */ False,
	/* destroy               */ NULL,
	/* resize                */ resize,
	/* expose                */ expose,
	/* set_values            */ set_values,
	/* set_values_hook       */ NULL,
	/* set_values_almost     */ XtInheritSetValuesAlmost,
	/* get_values_hook       */ NULL,
	/* accept_focus          */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* tm_table              */ XtInheritTranslations,
	/* query_geometry        */ query_geometry,
	/* display_accelerator   */ NULL,
	/* extension             */ (XtPointer)&_XmScrolledWCoreClassExtRec
    },
    /* Composite class part */
    {
	/* geometry manager */ geometry_manager, 
        /* change_managed   */ change_managed, 
        /* insert_child     */ insert_child,
        /* delete_child     */ delete_child,
        /* extension        */ NULL,
    },
    /* Constraint class part */
    {
	/* subresources      */ NULL,
        /* subresource_count */ 0,
        /* constraint_size   */ 0,
        /* initialize        */ NULL,
        /* destroy           */ NULL,
        /* set_values        */ NULL,
        /* extension         */ NULL,
    },
    /* XmManager class part */
    {
        /* translations                 */ _XmScrolledW_ScrolledWindowXlations,
        /* syn_resources                */ syn_resources,
        /* num_syn_resources            */ XtNumber(syn_resources),
        /* syn_constraint_resources     */ NULL,
        /* num_syn_constraint_resources */ 0,
        /* parent_process               */ XmInheritParentProcess,
        /* extension                    */ (XtPointer)&_XmScrolledWMClassExtRec
    },
    /* XmScrolledWindow part */
    {
	/* extension */ NULL,
    },
};
/* *INDENT-ON* */

WidgetClass xmScrolledWindowWidgetClass = (WidgetClass)&xmScrolledWindowClassRec;

#if 1
static void
class_initialize()
{
    _XmScrolledWCoreClassExtRec.record_type = XmQmotif;
}
#endif

static void
class_part_initialize(WidgetClass widget_class)
{
    XmScrolledWindowWidgetClass swclass =
	(XmScrolledWindowWidgetClass)widget_class;
    CompositeClassExtension ext, *extptr;

    extptr = (CompositeClassExtension *)
	_XmGetClassExtensionPtr((XmGenericClassExt *)
				&(swclass->composite_class.extension),
				NULLQUARK);

    if (extptr == NULL || *extptr == NULL)
    {
	ext = (CompositeClassExtension)XtNew(CompositeClassExtensionRec);
	if (ext != NULL)
	{
	    ext->next_extension = swclass->composite_class.extension;
	    ext->record_type = NULLQUARK;
	    ext->version = XtCompositeExtensionVersion;
	    ext->record_size = sizeof(CompositeClassExtensionRec);
	    ext->accepts_objects = True;
#if XtSpecificationRelease >= 6
	    ext->allows_change_managed_set = True;
#endif
	    swclass->composite_class.extension = (XtPointer)ext;
	}
    }
    _XmFastSubclassInit(widget_class, XmSCROLLED_WINDOW_BIT);
}

char *
GetRealTranslations(Widget w, XmTranslations tr)
{
    int buf_size = 100, len = 0, eventType;
    unsigned virt_keysym;
    KeySym keysym;
    unsigned int virt_modifiers, modifiers = 0;
    char *buf = XtMalloc(sizeof(char) * buf_size);
    char *key, map_not[] = "<MapNotify>: SWCWMapNotify()";

    buf[0] = '\0';
    for (;tr->virtkey != NULL; tr++)
    {
	if (_XmMapKeyEvent(tr->virtkey, &eventType,
			   &virt_keysym, &virt_modifiers))
	{
	    char mod_str[50];
	    int key_len;

	    _XmVirtualToActualKeysym(XtDisplay(w), virt_keysym,
				     &keysym, &modifiers);

	    modifiers |= virt_modifiers | tr->modifiers ;
	    key = XKeysymToString(keysym);
	    mod_str[0] = '\0';
	    if (modifiers & ShiftMask)
	    {
		strcat(mod_str, "Shift ");
	    }
	    if (modifiers & LockMask)
	    {
		strcat(mod_str, "Lock ");
	    }
	    if (modifiers & ControlMask)
	    {
		strcat(mod_str, "Ctrl ");
	    }
	    if (modifiers & Mod1Mask)
	    {
		strcat(mod_str, "Mod1 ");
	    }
 	    if (modifiers & Mod2Mask)
	    {
		strcat(mod_str, "Mod2 ");
	    }
 	    if (modifiers & Mod3Mask)
	    {
		strcat(mod_str, "Mod3 ");
	    }
 	    if (modifiers & Mod4Mask)
	    {
		strcat(mod_str, "Mod4 ");
	    }
 	    if (modifiers & Mod5Mask)
	    {
		strcat(mod_str, "Mod5 ");
	    }
	    strcat(mod_str, "<Key>");
	    key_len = strlen(mod_str) + strlen(key) +
		strlen(tr->function) + 3;
	    if (buf_size < key_len + len + 1)
	    {
		buf_size += buf_size;
		buf = XtRealloc(buf, sizeof(char) * buf_size);
	    }
	    strcat(buf, mod_str);
	    strcat(buf, key);
	    strcat(buf,": ");
	    strcat(buf, tr->function);
	    strcat(buf,"\n");
	    len += key_len;
	}
    }

    if (buf_size < strlen(map_not) + len + 1)
    {
	buf_size += buf_size;
	buf = XtRealloc(buf, sizeof(char) * buf_size);
    }
    strcat(buf, map_not);
    return buf;
}

/*
 * Note
 * The OSF/Motif manual page for XmScrolledWindow describes a "clip" widget
 * and a "work" widget. The Clip widget is really the area that the user sees,
 * whereas the Work widget is the one that is put in the ScrolledWindow.
 * I.e. if you display a large label widget, then that label is the work widget,
 * and the clip widget is something magic that is private to scrolledwindow.
 */
static void
initialize(Widget request, Widget new_w,
	   ArgList args, Cardinal *num_args)
{
    SW_InInit(new_w) = True;
    if (!XmRepTypeValidValue(XmRepTypeGetId(XmRScrollingPolicy),
			     SW_ScrollPolicy(new_w),
			     new_w))
    {
	SW_ScrollPolicy(new_w) = XmAPPLICATION_DEFINED;
    }

    if (!XmRepTypeValidValue(XmRepTypeGetId(XmRVisualPolicy),
			     SW_VisualPolicy(new_w),
			     new_w))
    {
	SW_VisualPolicy(new_w) = XmVARIABLE;
    }

    if (!XmRepTypeValidValue(XmRepTypeGetId(XmRScrollBarPlacement),
			     SW_Placement(new_w),
			     new_w))
    {
	SW_Placement(new_w) = XmBOTTOM_RIGHT;
    }

    if (SW_ScrollPolicy(new_w) == XmAPPLICATION_DEFINED)
    {
	if (SW_VisualPolicy(new_w) == XmCONSTANT)
	{
	    	_XmWarning(new_w, _XmMsgScrolledW_0009);
	}
	SW_VisualPolicy(new_w) = XmVARIABLE;
    }
    else
    {
	SW_VisualPolicy(new_w) = XmCONSTANT;
    }

    if (SW_VisualPolicy(new_w) == XmVARIABLE)
    {
	if (SW_ScrollBarPolicy(new_w) == XmAS_NEEDED)
	{
	    	_XmWarning(new_w, _XmMsgScrolledW_0006);
	}
	SW_ScrollBarPolicy(new_w) = XmSTATIC;
    }
    else
    {
	if (SW_ScrollBarPolicy(new_w) == (unsigned char)XmUNSPECIFIED ||
	    !XmRepTypeValidValue(XmRepTypeGetId(XmRScrollBarDisplayPolicy),
				 SW_ScrollBarPolicy(new_w),
				 new_w))
	{
	    SW_ScrollBarPolicy(new_w) = XmAS_NEEDED;
	}
    }


    if (SW_Spacing(new_w) == XmINVALID_DIMENSION)
    {
	SW_Spacing(new_w) = 4;
    }

    if (MGR_ShadowThickness(new_w) == XmINVALID_DIMENSION)
    {
	if (SW_ScrollPolicy(new_w) == XmAUTOMATIC)
	{
	    MGR_ShadowThickness(new_w) = 2;
	}
	else
	{
	    MGR_ShadowThickness(new_w) = 0;
	}
    }

    if (SW_VisualPolicy(new_w) == XmCONSTANT)
    {
	char *s;

	SW_ClipWindow(new_w) =
	    (XmDrawingAreaWidget)
	    XtVaCreateManagedWidget("clipWindow",
				    xmDrawingAreaWidgetClass, new_w,
				    XmNresizePolicy, XmRESIZE_SWINDOW,
				    NULL);
	XtRegisterGrabAction(BeginLineGrab, True, 
			     KeyPressMask | KeyReleaseMask,
			     GrabModeAsync,
			     GrabModeAsync);
	XtRegisterGrabAction(EndLineGrab, True, 
			     KeyPressMask | KeyReleaseMask,
			     GrabModeAsync,
			     GrabModeAsync);
	XtRegisterGrabAction(BeginDataGrab, True, 
			     KeyPressMask | KeyReleaseMask,
			     GrabModeAsync,
			     GrabModeAsync);
	XtRegisterGrabAction(EndDataGrab, True, 
			     KeyPressMask | KeyReleaseMask,
			     GrabModeAsync,
			     GrabModeAsync);
	XtRegisterGrabAction(PageLeftGrab, True, 
			     KeyPressMask | KeyReleaseMask,
			     GrabModeAsync,
			     GrabModeAsync);
	XtRegisterGrabAction(PageRightGrab, True, 
			     KeyPressMask | KeyReleaseMask,
			     GrabModeAsync,
			     GrabModeAsync);
	XtRegisterGrabAction(PageUpGrab, True, 
			     KeyPressMask | KeyReleaseMask,
			     GrabModeAsync,
			     GrabModeAsync);
	XtRegisterGrabAction(PageDownGrab, True, 
			     KeyPressMask | KeyReleaseMask,
			     GrabModeAsync,
			     GrabModeAsync);
	/* 
	 * This sucker comes from a incompatibility between virtual keys
	 * and GrabAction. It is necessary to bind keys both to the
	 * real keysyms and to the virtual keysyms. And to make things 
	 * even worse it will only work if the virtual keysyms is bound
	 * after the widget is realized.
	 */

	if (RealTranslationsParsed == NULL)
	{
	    s = GetRealTranslations(new_w, ClipWindowTranslations);
	    RealTranslationsParsed = XtParseTranslationTable(s);
	    XtFree(s);
	}
	if (TranslationsParsed == NULL)
	{
	TranslationsParsed =
	    XtParseTranslationTable(_XmScrolledW_ClipWindowTranslationTable);
	}

	XtOverrideTranslations((Widget) SW_ClipWindow(new_w), 
			       RealTranslationsParsed);


	SW_VSB(new_w) =
	    (XmScrollBarWidget)XtVaCreateWidget("VertScrollBar",
						xmScrollBarWidgetClass, new_w,
						XmNorientation, XmVERTICAL,
						XmNsliderSize, 100,
						XmNforeground,
						 MGR_Foreground(new_w),
						XmNbackground,
						 XtBackground(new_w),
						XmNtopShadowColor,
						 MGR_TopShadowColor(new_w),
						XmNtopShadowPixmap,
						 MGR_TopShadowPixmap(new_w),
						XmNbottomShadowColor,
						 MGR_BottomShadowColor(new_w),
						XmNbottomShadowPixmap,
						 MGR_BottomShadowPixmap(new_w),
						NULL);
	XtAddCallback((Widget)SW_VSB(new_w),
		      XmNincrementCallback,
		      VsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_VSB(new_w),
		      XmNdecrementCallback,
		      VsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_VSB(new_w),
		      XmNpageIncrementCallback,
		      VsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_VSB(new_w),
		      XmNpageDecrementCallback,
		      VsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_VSB(new_w),
		      XmNdragCallback,
		      VsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_VSB(new_w),
		      XmNvalueChangedCallback,
		      VsbCallback,
		      NULL);
	SW_HasVSB(new_w) = True;
	XtManageChild((Widget) SW_VSB(new_w));


	SW_HSB(new_w) =
	    (XmScrollBarWidget)XtVaCreateWidget("HorScrollBar",
						xmScrollBarWidgetClass, new_w,
						XmNorientation, XmHORIZONTAL,
						XmNsliderSize, 100,
						XmNforeground,
						 MGR_Foreground(new_w),
						XmNbackground,
						 XtBackground(new_w),
						XmNtopShadowColor,
						 MGR_TopShadowColor(new_w),
						XmNtopShadowPixmap,
						 MGR_TopShadowPixmap(new_w),
						XmNbottomShadowColor,
						 MGR_BottomShadowColor(new_w),
						XmNbottomShadowPixmap,
						 MGR_BottomShadowPixmap(new_w),
						NULL);

	XtAddCallback((Widget)SW_HSB(new_w),
		      XmNincrementCallback,
		      HsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_HSB(new_w),
		      XmNdecrementCallback,
		      HsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_HSB(new_w),
		      XmNpageIncrementCallback,
		      HsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_HSB(new_w),
		      XmNpageDecrementCallback,
		      HsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_HSB(new_w),
		      XmNdragCallback,
		      HsbCallback,
		      NULL);
	XtAddCallback((Widget)SW_HSB(new_w),
		      XmNvalueChangedCallback,
		      HsbCallback,
		      NULL);
	XtManageChild((Widget) SW_HSB(new_w));
	SW_HasHSB(new_w) = True;
    }
    else
    {
	SW_HasHSB(new_w) = False;
	SW_HasVSB(new_w) = False;
	SW_WorkWindow(new_w) = NULL;
	SW_ClipWindow(new_w) = NULL;
	SW_HSB(new_w) = NULL;
	SW_VSB(new_w) = NULL;

    }

    SW_VSBMinimum(new_w) = 0;
    SW_VSBMaximum(new_w) = 0;
    SW_VSBValue(new_w) = 0;
    SW_VSBSliderSize(new_w) = 0;

    SW_HSBMinimum(new_w) = 0;
    SW_HSBMaximum(new_w) = 0;
    SW_HSBValue(new_w) = 0;
    SW_HSBSliderSize(new_w) = 0;

    SW_HSBX(new_w) = 0;
    SW_HSBY(new_w) = 0;
    SW_HSBWidth(new_w) = 0;
    SW_HSBHeight(new_w) = 0;

    SW_VSBX(new_w) = 0;
    SW_VSBY(new_w) = 0;
    SW_VSBWidth(new_w) = 0;
    SW_VSBHeight(new_w) = 0;

    /*
     * MLM -- 60 and 20 I got from comparing with testXm/scrolledw/test1.motif
     * these do appear to be default Motif values.
     *
     * Danny - Trust me, they're supposed to be 100x100.
     *
     * rws 20 Jul 1997
     * Does this really need to be here?? Taking this out lets me avoid the
     * XmIsScrolledWindow junk in Form:constraint_initialize.  I do not think
     * this hurts anything in the tests or any apps.
     *
     * MLM: I have my suspicions about this.  I think what may happen is that
     * query_geometry returns 100x100 iff w/h is 0x0.
     *
     * Once again, Danny's right.  However, adding the ScrollPolicy check
     * makes all the SW tests appear correctly.
     *
     * HOWEVER, if it isn't defined, then xmgr works right (initial display)
     */
    SW_GivenWidth(new_w) = XtWidth(request);
    SW_GivenHeight(new_w) = XtHeight(request);

    if (SW_GivenWidth(new_w) == 0)
    {
	XtWidth(new_w) = 100;
    }
    if (SW_GivenHeight(new_w) == 0)
    {
	XtHeight(new_w) = 100;
    }
    SW_XOffset(new_w) = SW_MarginHeight(new_w);
    SW_YOffset(new_w) = SW_MarginWidth(new_w);


    SW_FromResize(new_w) = False;


    /* There's something spooky about highlighting here :
     * Scrollbars are not highlighted by default except when ...
     * they're in an XmAUTOMATIC scrolledwindow.
     * Surprise :-)
     */
    if (SW_ScrollPolicy(new_w) == XmAUTOMATIC)
    {
	if (Prim_HighlightThickness(SW_VSB(new_w)) == 0)
	{
	    Prim_HighlightThickness(SW_VSB(new_w)) = 2;
	    XtWidth(SW_VSB(new_w)) += 4;
	}
	if (Prim_HighlightThickness(SW_HSB(new_w)) == 0)
	{
	    Prim_HighlightThickness(SW_HSB(new_w)) = 2;
	    XtHeight(SW_HSB(new_w)) += 4;
	}
    }
    SW_InInit(new_w) = False;

}

/* Give the preferred size with the given constraints
 */

void
_XmScrolledWindowPreferredSize(Widget w, XtWidgetGeometry *request,
			  XtWidgetGeometry *answer)
{
    Dimension work_height, work_width, usable_width, usable_height, sw_st;
    Boolean show_vsb, show_hsb;
    Dimension vsb_w, hsb_h, vsb_ht, hsb_ht;
    
    answer->request_mode =  CWWidth | CWHeight;
    if (SW_WorkWindow(w) == NULL)
    {
	answer->width = XtWidth(w);
	answer->height = XtHeight(w);
    }
    else
    {
	usable_width = XtWidth(w) - SW_XOffset(w) - SW_MarginWidth(w);
	usable_height = XtHeight(w) - SW_YOffset(w) - SW_MarginHeight(w);
	sw_st = MGR_ShadowThickness(w);
	show_vsb = SW_VSB(w) != NULL && XtIsManaged(SW_VSB(w));
	show_hsb = SW_HSB(w) != NULL && XtIsManaged(SW_HSB(w));
	vsb_w = show_vsb ? XtWidth(SW_VSB(w)): 0;
	hsb_h = show_hsb ? XtHeight(SW_HSB(w)): 0;
	vsb_ht = show_vsb ? Prim_HighlightThickness(SW_VSB(w)): 0;
	hsb_ht = show_hsb ? Prim_HighlightThickness(SW_HSB(w)): 0;
	work_width = XtWidth(SW_WorkWindow(w)) +
	    2 * XtBorderWidth(SW_WorkWindow(w));
	work_height = XtHeight(SW_WorkWindow(w)) +
	    2 * XtBorderWidth(SW_WorkWindow(w));

	if ((request->request_mode & (CWWidth | CWHeight)) == CWWidth )
	{
	    answer->width = request->width;

	    if (SW_ScrollBarPolicy(w) == XmSTATIC ||
	        (SW_ScrollBarPolicy(w) == XmAS_NEEDED &&
		 request->width < work_width + 2 * sw_st))
	    {
		show_hsb &= True;
	    }
	    if (SW_ScrollBarPolicy(w) == XmAS_NEEDED)
	    {
		show_vsb = False;
	    }
	    if (show_hsb && show_vsb)
	    {
		answer->height = work_height + 
		    2 * (sw_st + vsb_ht) +
		    SW_Spacing(w) + hsb_h + SW_YOffset(w) +
		    SW_MarginHeight(w);
	    }
	    else if (show_hsb)
	    {
		answer->height = work_height + 
		    2 * (sw_st + vsb_ht) + SW_Spacing(w) + hsb_h + 
		    SW_YOffset(w) + SW_MarginHeight(w);
	    }
	    else if (show_vsb)
	    {
		answer->height = work_height + 
		    2 * (sw_st + vsb_ht) +
		    SW_YOffset(w) + SW_MarginHeight(w);
	    }
	    else
	    {
		answer->height = work_height + 
		    2 * sw_st + SW_YOffset(w) + SW_MarginHeight(w);
	    }
	}
	else if ((request->request_mode & (CWWidth | CWHeight)) ==  CWHeight)
	{
	    answer->height = request->height;

	    if (SW_ScrollBarPolicy(w) == XmSTATIC ||
	        (SW_ScrollBarPolicy(w) == XmAS_NEEDED &&
		 request->height < work_height + 2 * sw_st))
	    {
		show_vsb &= True;
	    }
	    if (SW_ScrollBarPolicy(w) == XmAS_NEEDED)
	    {
		show_hsb = False;
	    }
	    if (show_hsb && show_vsb)
	    {
		answer->width = work_width + 
		    2 * (sw_st + hsb_ht) +
		    SW_Spacing(w) + vsb_w + SW_XOffset(w) +
		    SW_MarginWidth(w);
	    }
	    else if (show_vsb)
	    {
		answer->width = work_width + 
		    2 * (sw_st + hsb_ht) + SW_Spacing(w) + vsb_w + 
		    SW_XOffset(w) + SW_MarginWidth(w);
	    }
	    else if (show_hsb)
	    {
		answer->width = work_width + 
		    2 * (sw_st + hsb_ht) +
		    SW_XOffset(w) + SW_MarginWidth(w);
	    }
	    else
	    {
		answer->width = work_width + 
		    2 * sw_st + SW_XOffset(w) + SW_MarginWidth(w);
	    }
	}
	else
	{
	    if (SW_ScrollBarPolicy(w) == XmAS_NEEDED)
	    {
		show_hsb = False;
		show_vsb = False;
	    }
	    if (show_hsb && show_vsb)
	    {
		answer->width = work_width + 
		    2 * (sw_st + hsb_ht) +
		    SW_Spacing(w) + vsb_w + SW_XOffset(w) +
		    SW_MarginWidth(w);
		answer->height = work_height + 
		    2 * (sw_st + vsb_ht) +
		    SW_Spacing(w) + hsb_h + SW_YOffset(w) +
		    SW_MarginHeight(w);
	    }
	    else if (show_vsb)
	    {
		answer->width = work_width + 
		    2 * (sw_st + hsb_ht)+ SW_Spacing(w) + vsb_w + 
		    SW_XOffset(w) + SW_MarginWidth(w);
		answer->height = work_height + 
		    2 * (sw_st + vsb_ht) +
		    SW_YOffset(w) + SW_MarginHeight(w);
	    }
	    else if (show_hsb)
	    {
		answer->width = work_width + 
		    2 * (sw_st + hsb_ht) +
		    SW_XOffset(w) + SW_MarginWidth(w);
		answer->height = work_height + 
		    2 * (sw_st + vsb_ht) + SW_Spacing(w) + hsb_h + 
		    SW_YOffset(w) + SW_MarginHeight(w);
	    }
	    else
	    {
		answer->width = work_width + 
		    2 * sw_st + SW_XOffset(w) + SW_MarginWidth(w);
		answer->height = work_height + 
		    2 * sw_st + SW_YOffset(w) + SW_MarginHeight(w);
	    }
	}
    }
}


/* Try to resize to a size where the workarea and scrollbars fits into the
 * window
 */
static void
MakeResizeRequest(Widget w)
{
    XtWidgetGeometry request, answer;
    Dimension width, new_width, height, new_height;


    if (SW_WorkWindow(w) == NULL || !XtIsManaged(SW_WorkWindow(w)))
    {
	return;
    }
    request.request_mode = 0;
    _XmScrolledWindowPreferredSize(w, &request, &answer);
    width = answer.width;
    height = answer.height;

    if (!XtIsRealized(w))
    {
	if (SW_GivenWidth(w) != 0)
	{
	    width = SW_GivenWidth(w);
	}
	if (SW_GivenHeight(w) != 0)
	{
	    height = SW_GivenHeight(w);
	}
    }


    if (XtMakeResizeRequest(w, width, height, &new_width, &new_height) ==
	XtGeometryAlmost)
    {
	XtMakeResizeRequest(w, new_width, new_height, &new_width, &new_height);
    }
}


static Boolean
set_values(Widget old, Widget request, Widget new_w,
	   ArgList args, Cardinal *num_args)
{
    Boolean r = False;		/* Must I be redisplayed ? */

    DEBUGOUT(XdbDebug(__FILE__, new_w,
		      "set_values: %i args\n"
		      "\t    old X %5i Y %5i W %5i H %5i\n"
		      "\trequest X %5i Y %5i W %5i H %5i\n"
		      "\t  new_w X %5i Y %5i W %5i H %5i\n",
		      *num_args,
		      XtX(old), XtY(old),
		      XtWidth(old), XtHeight(old),
		      XtX(request), XtY(request),
		      XtWidth(request), XtHeight(request),
		      XtX(new_w), XtY(new_w),
		      XtWidth(new_w), XtHeight(new_w)));
    DEBUGOUT(XdbPrintArgList(__FILE__, new_w, args, *num_args, False));

#define	NE(x)	(x(new_w) != x(old))

    if (NE(SW_ClipWindow))
    {
	SW_ClipWindow(new_w) = SW_ClipWindow(old);
	_XmWarning(new_w,
		   "Attempted to change the clipWindow in scrolled window %s",
		   XtName(new_w));
    }

    if (NE(SW_VisualPolicy))
    {
	SW_VisualPolicy(new_w) = SW_VisualPolicy(old);
	_XmWarning(new_w, _XmMsgScrolledW_0004);
    }

    if (NE(SW_ScrollPolicy))
    {
	SW_ScrollPolicy(new_w) = SW_ScrollPolicy(old);
	_XmWarning(new_w,
	       "Attempted to change the scrollingPolicy in scrolled window %s",
		   XtName(new_w));
    }

    if (NE(SW_HSB))
    {
	if (SW_HSB(new_w) == NULL)
	{
	    SW_HSB(new_w) = SW_HSB(old);
	}
	else if (SW_ScrollPolicy(new_w) == XmAUTOMATIC)
	{
	    _XmWarning(new_w,_XmMsgScrolledW_0007);
	    SW_HSB(new_w) = SW_HSB(old);
	}
	else
	{
	    if (SW_HSB(old) != NULL)
	    {
		XtUnmanageChild((Widget) SW_HSB(old));
	    }
	    r = True;
	}
    }
    if (NE(SW_VSB))
    {
	if (SW_VSB(new_w) == NULL)
	{
	    SW_VSB(new_w) = SW_VSB(old);
	}
	else if (SW_ScrollPolicy(new_w) == XmAUTOMATIC)
	{
	    _XmWarning(new_w,_XmMsgScrolledW_0007);
	    SW_VSB(new_w) = SW_VSB(old);
	}
	else
	{
	    if (SW_VSB(old) != NULL)
	    {
		XtUnmanageChild((Widget) SW_VSB(old));
	    }
	    r = True;
	}
    }

    if (NE(SW_ScrollBarPolicy))
    {
	if (!XmRepTypeValidValue(XmRepTypeGetId(XmRScrollBarDisplayPolicy),
				 SW_ScrollBarPolicy(new_w),
				 new_w))
	{
	    SW_ScrollPolicy(new_w) = SW_ScrollPolicy(old);
	}
	else if (SW_ScrollBarPolicy(new_w) == XmAS_NEEDED &&
		 SW_VisualPolicy(new_w) == XmVARIABLE)
	{
	    _XmWarning(new_w, _XmMsgScrolledW_0005);
	    SW_ScrollBarPolicy(new_w) = XmSTATIC;
	}
	else
	{
	    r = True;
	}
    }
    if (NE(SW_MarginHeight))
    {
	SW_XOffset(new_w) = SW_MarginHeight(new_w);
	r = True;
    }
    if (NE(SW_MarginWidth))
    {
	SW_YOffset(new_w) = SW_MarginWidth(new_w);
	r = True;
    }
    if (NE(SW_Placement) || NE(SW_Spacing))
    {
	r = True;
    }

    if (NE(SW_WorkWindow))
    {
	DEBUGOUT(XdbDebug0(__FILE__, new_w,
			   "Work Window changed: from %s to %s.\n",
			   SW_WorkWindow(old) != NULL
				? XtName(SW_WorkWindow(old))
				: "(null)",
			   SW_WorkWindow(new_w) != NULL
				? XtName(SW_WorkWindow(new_w))
				: "(null)"));
	DEBUGOUT(XdbDebug0(__FILE__, new_w,
			   "Work Window was: %dx%d now %dx%d.\n",
			   SW_WorkWindow(old) != NULL
				? XtWidth(SW_WorkWindow(old))
				: 0,
			   SW_WorkWindow(old) != NULL
				? XtHeight(SW_WorkWindow(old))
				: 0,
			   SW_WorkWindow(new_w) != NULL
				? XtWidth(SW_WorkWindow(new_w))
				: 0,
			   SW_WorkWindow(new_w) != NULL
				? XtHeight(SW_WorkWindow(new_w))
				: 0));
	DEBUGOUT(XdbDebug0(__FILE__, new_w, "... SW_VisualPolicy %s",
			   (SW_VisualPolicy(new_w) == XmCONSTANT)
				? "XmCONSTANT"
				: (SW_VisualPolicy(new_w) == XmVARIABLE)
					? "XmVARIABLE"
					: (SW_VisualPolicy(new_w) ==
					   XmRESIZE_IF_POSSIBLE)
						? "XmRESIZE_IF_POSSIBLE"
						: "???"));
	DEBUGOUT(XdbDebug0(__FILE__, new_w, " SW_ScrollBarPolicy %s",
			   (SW_ScrollBarPolicy(new_w) == XmSTATIC)
			   ? "XmSTATIC"
			   : (SW_ScrollBarPolicy(new_w) == XmAS_NEEDED)
			   ? "XmAS_NEEDED"
			   : "???"));
	DEBUGOUT(XdbDebug0(__FILE__, new_w, " SW_ScrollPolicy %s\n",
			   (SW_ScrollPolicy(new_w) == XmAPPLICATION_DEFINED)
			   ? "XmAPPLICATION_DEFINED"
			   : (SW_ScrollPolicy(new_w) == XmAUTOMATIC)
			   ? "XmAUTOMATIC"
			   : "???"));
	/* rws 6 Jun 1998
	   If we change the work area we definitly need to refresh the
	   display.  This showed as 1x1 <textarea>'s in Mozilla
	 */
	if (SW_WorkWindow(old) != NULL)
	{
	    XtUnmanageChild(SW_WorkWindow(old));
	}
	if (XtIsManaged(SW_WorkWindow(new_w)))
	{
	    r = True;
	}
    }

    if (XtWidth(new_w) != XtWidth(old))
    {
	SW_GivenWidth(new_w) = XtWidth(new_w);
    }
    if (XtHeight(new_w) != XtHeight(old))
    {
	SW_GivenHeight(new_w) = XtHeight(new_w);
    }
    if (NE(SW_WorkWindow) && SW_VisualPolicy(new_w) == XmVARIABLE)
    {
	MakeResizeRequest(new_w);
    }
    
    /* init chained from super to sub order */
    /* rws 6 Jun 1998
       If we are not realized do not do all of this geo stuff.  It will be
       done when we get realized.  This ties in with the Mozilla and ml stuff
       above.
     */
    if (r && XtIsRealized(new_w))
    {
	(*XtClass(new_w)->core_class.resize) (new_w);
    }

    return r;
}

static void
VariableLayout(Widget w, XmSWValues *current, XmSWValues *new)
{
    Dimension usable_width, usable_height;
    Dimension vsb_ht, hsb_ht, sw_st;
    Boolean too_small = False, has_work;

    *new = *current;
    has_work = SW_WorkWindow(w) != NULL && XtIsManaged(SW_WorkWindow(w));
    usable_width = new->SwW - SW_XOffset(w) - SW_MarginWidth(w);
    usable_height = new->SwH - SW_YOffset(w) - SW_MarginHeight(w);
    hsb_ht = new->HasHSB ? Prim_HighlightThickness(SW_HSB(w)) : 0;
    vsb_ht = new->HasVSB ? Prim_HighlightThickness(SW_VSB(w)) : 0;
    sw_st = MGR_ShadowThickness(w);
    new->ClipW = 0;
    new->ClipH = 0;
    new->ClipX = 0;
    new->ClipY = 0;

    /*
     * Check if it is possible to fit the widgets in the remaining area,
     * and if not place them outside the visible area
     */
    if (new->HasHSB && new->HasVSB)
    {
	if (usable_width <= new->VsbW + SW_Spacing(w) + 2 * (hsb_ht + sw_st) ||
	    usable_height <= new->HsbH + SW_Spacing(w) + 2 * (vsb_ht + sw_st))
	    too_small = True;
    }
    else if (new->HasHSB)
    {
	if (usable_width <= 2 * (hsb_ht + sw_st) ||
	    usable_height <= new->HsbH + SW_Spacing(w) + 2 *sw_st)
	    too_small = True;
    }
    else if (new->HasVSB)
    {
	if (usable_width <= new->VsbW + SW_Spacing(w) + 2 *sw_st ||
	    usable_height <= 2 * (vsb_ht + sw_st))
	    too_small = True;
    }
    else
    {
	if (usable_width <= 2 *sw_st || usable_height <= 2 *sw_st)
	    too_small = True;
    }
    if (too_small)
    {
	new->ShowHSB = False;
	new->HsbX = new->SwW;
	new->HsbY = 0;
	new->ShowVSB = False;
	new->VsbX = new->SwW;
	new->VsbY = 0;
	new->WorkX = new->SwW + sw_st;
	new->WorkY = 0;
    }
    else
    {
	if (new->ShowHSB && new->ShowVSB)
	{
	    new->HsbW = usable_width - new->VsbW - SW_Spacing(w);
	    new->VsbH = usable_height - new->HsbH - SW_Spacing(w);
	    new->WorkW = new->HsbW - 2 * (hsb_ht + sw_st);
	    new->WorkH = new->VsbH - 2 * (vsb_ht + sw_st);
	}
	else if (new->ShowHSB)
	{
	    new->HsbW = usable_width;
	    new->VsbH = usable_height - new->HsbH - SW_Spacing(w);
	    new->WorkW = usable_width - 2 * sw_st;
	    new->WorkH = new->VsbH - 2 * (vsb_ht + sw_st);
	}
	else if (new->ShowVSB)
	{
	    new->HsbW = usable_width - new->VsbW - SW_Spacing(w);
	    new->VsbH = usable_height;
	    new->WorkW = new->HsbW - 2 * (hsb_ht + sw_st);
	    new->WorkH = usable_height - 2 * sw_st;
	}
	else
	{
	    new->HsbW = usable_width;
	    new->VsbH = usable_height;
	    new->WorkW = usable_width - 2 * sw_st;
	    new->WorkH = usable_height - 2 * sw_st;
	}
	switch(SW_Placement(w))
	{
	case XmTOP_LEFT:
	    if (new->ShowHSB && new->ShowVSB)
	    {
		new->HsbX = SW_XOffset(w) + new->VsbW + SW_Spacing(w);
		new->HsbY = SW_YOffset(w);
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w) + new->HsbH + SW_Spacing(w);
		new->WorkX = new->HsbX + hsb_ht + sw_st;
		new->WorkY = new->VsbY + vsb_ht + sw_st;
	    }
	    else if (new->ShowHSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w);
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->WorkX = new->HsbX + hsb_ht + sw_st;
		new->WorkY = SW_YOffset(w) + new->HsbH + SW_Spacing(w) + vsb_ht + sw_st;
	    }
	    else if (new->ShowVSB)
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w);
		new->WorkX = SW_XOffset(w) + new->VsbW + SW_Spacing(w) + hsb_ht + sw_st;
		new->WorkY = new->VsbY + vsb_ht + sw_st;
	    }
	    else
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->WorkX = SW_XOffset(w) + sw_st;
		new->WorkY = SW_YOffset(w) + sw_st;
	    }
	    break;
	case XmBOTTOM_LEFT:
	    if (new->ShowHSB && new->ShowVSB)
	    {
		new->HsbX = SW_XOffset(w) + new->VsbW + SW_Spacing(w);
		new->HsbY = SW_YOffset(w) + new->VsbH + SW_Spacing(w);
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w);
		new->WorkX = new->HsbX + hsb_ht + sw_st;
		new->WorkY = new->VsbY + vsb_ht + sw_st;
	    }
	    else if (new->ShowHSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w) + usable_height - new->HsbH;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->WorkX = new->HsbX + hsb_ht + sw_st;
		new->WorkY = SW_YOffset(w) + sw_st;
	    }
	    else if (new->ShowVSB)
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w);
		new->WorkX = SW_XOffset(w) + sw_st;
		new->WorkY = new->VsbY + vsb_ht + sw_st;
	    }
	    else
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->WorkX = SW_XOffset(w) + sw_st;
		new->WorkY = SW_YOffset(w) + sw_st;
	    }
	    break;
	case XmTOP_RIGHT:
	    if (new->ShowHSB && new->ShowVSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w);
		new->VsbX = SW_XOffset(w) + usable_width - new->VsbW;
		new->VsbY = SW_YOffset(w) + new->HsbH + SW_Spacing(w);
		new->WorkX = new->HsbX + hsb_ht + sw_st;
		new->WorkY = new->VsbY + vsb_ht + sw_st;
	    }
	    else if (new->ShowHSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w);
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->WorkX = new->HsbX + hsb_ht + sw_st;
		new->WorkY = SW_YOffset(w) + new->HsbH + SW_Spacing(w) + vsb_ht + sw_st;
	    }
	    else if (new->ShowVSB)
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w) + usable_width - new->VsbW;
		new->WorkX = SW_XOffset(w) + sw_st;
		new->WorkY = new->VsbY + vsb_ht + sw_st;
	    }
	    else
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->WorkX = SW_XOffset(w) + sw_st;
		new->WorkY = SW_YOffset(w) + sw_st;
	    }
	    break;
	case XmBOTTOM_RIGHT:
	    if (new->ShowHSB && new->ShowVSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w) + usable_height - new->HsbH;
		new->VsbX = SW_XOffset(w) + usable_width - new->VsbW;
		new->VsbY = SW_YOffset(w);
		new->WorkX = new->HsbX + hsb_ht + sw_st;
		new->WorkY = new->VsbY + vsb_ht + sw_st;
	    }
	    else if (new->ShowHSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w) + usable_height - new->HsbH;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->WorkX = new->HsbX + hsb_ht + sw_st;
		new->WorkY = SW_YOffset(w) + sw_st;
	    }
	    else if (new->ShowVSB)
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = SW_XOffset(w) + usable_width - new->VsbW;
		new->VsbY = SW_YOffset(w);
		new->WorkX = SW_XOffset(w) + sw_st;
		new->WorkY = new->VsbY + vsb_ht + sw_st;
	    }
	    else
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->WorkX = SW_XOffset(w) + sw_st;
		new->WorkY = SW_YOffset(w) + sw_st;
	    }
	    break;
	default:
	}
    }
}







static void
ConstantLayout(Widget w, XmSWValues *current, XmSWValues *new)
{
    Dimension usable_width, usable_height;
    Dimension vsb_ht, hsb_ht, sw_st;
    Boolean too_small = False, has_clip, has_work;

    *new = *current;

    has_clip = SW_ClipWindow(w) != NULL && XtIsManaged(SW_ClipWindow(w));
    has_work = SW_WorkWindow(w) != NULL && XtIsManaged(SW_WorkWindow(w));
    usable_width = new->SwW - SW_XOffset(w) - SW_MarginWidth(w);
    usable_height = new->SwH - SW_YOffset(w) - SW_MarginHeight(w);
    hsb_ht = new->HasHSB ? Prim_HighlightThickness(SW_HSB(w)) : 0;
    vsb_ht = new->HasVSB ? Prim_HighlightThickness(SW_VSB(w)) : 0;
    sw_st = MGR_ShadowThickness(w);

    /* Check if it is possible to fit the widgets in the remaining area,
     * and if not place them outside the visible area
     */
    if (SW_ScrollBarPolicy(w) == XmSTATIC ||
	new->WorkW > usable_width - 2 * sw_st ||
	new->WorkH > usable_height - 2 * sw_st)
    {
	if (new->HasHSB && new->HasVSB)
	{
	    if (usable_width <= new->VsbW + SW_Spacing(w) + 2 * (hsb_ht + sw_st) ||
		usable_height <= new->HsbH + SW_Spacing(w) + 2 *(vsb_ht + sw_st))
		too_small = True;
	}
	else if (new->HasHSB)
	{
	    if (usable_width <= 2 * (hsb_ht + sw_st) ||
		usable_height <= new->HsbH + SW_Spacing(w) + 2 * sw_st)
		too_small = True;
	}
	else if (new->HasVSB)
	{
	    if (usable_width <= new->VsbW + SW_Spacing(w) + 2 * sw_st ||
		usable_height <= 2 * (vsb_ht + sw_st))
		too_small = True;
	}
	else
	{
	    if (usable_width <= 2 * sw_st || usable_height <= 2 *sw_st)
		too_small = True;
	}
    }
    if (too_small)
    {
	new->ShowHSB = False;
	new->HsbX = new->SwW;
	new->HsbY = 0;
	new->ShowVSB = False;
	new->VsbX = new->SwW;
	new->VsbY = 0;
	new->ClipX = new->SwW + sw_st;
	new->ClipY = 0;
    }
    else
    {
	new->ShowHSB = new->HasHSB;
	new->ShowVSB = new->HasVSB;
	if (SW_ScrollBarPolicy(w) == XmAS_NEEDED)
	{
	    if (new->WorkW <= usable_width - 2 * sw_st &&
		new->WorkH <= usable_height - 2* sw_st)
	    {
		new->ShowHSB = False;
		new->ShowVSB = False;
	    }
	    else if (new->WorkW > usable_width - 2 * sw_st &&
		     new->WorkH <= usable_height - 2 * sw_st - new->HsbH - 
		     SW_Spacing(w) - vsb_ht)
	    {
		new->ShowVSB = False;
	    }
	    else if (new->WorkH > usable_height - 2 * sw_st &&
		     new->WorkW <= usable_width - 2 * sw_st - new->VsbW - 
		     SW_Spacing(w) - hsb_ht)
	    {
		new->ShowHSB = False;
	    }
	}
	if (new->ShowHSB && new->ShowVSB)
	{
	    new->HsbW = usable_width - new->VsbW - SW_Spacing(w);
	    new->VsbH = usable_height - new->HsbH - SW_Spacing(w);
	    new->ClipW = new->HsbW - 2 * (hsb_ht + sw_st);
	    new->ClipH = new->VsbH - 2 * (vsb_ht + sw_st);
	}
	else if (new->ShowHSB)
	{
	    new->HsbW = usable_width;
	    new->ClipW = new->HsbW - 2 * (hsb_ht + sw_st);
	    new->ClipH = usable_height - new->HsbH - SW_Spacing(w) -
		vsb_ht - 2 * sw_st;
	}
	else if (new->ShowVSB)
	{
	    new->VsbH = usable_height;
	    new->ClipW = usable_width - new->VsbW - SW_Spacing(w) -
		hsb_ht - 2 * sw_st;
	    new->ClipH = new->VsbH - 2 * (vsb_ht + sw_st);
	}
	else
	{
	    new->ClipW = usable_width - 2* sw_st;
	    new->ClipH = usable_height - 2 * sw_st;
	}
	if (new->WorkW <= new->ClipW)
	{
	    new->WorkX = 0;
	}
	else if (new->WorkX + new->WorkW < new->ClipW)
	{
	    new->WorkX = new->ClipW - new->WorkW;
	}
	if (new->WorkH <= new->ClipH)
	{
	    new->WorkY = 0;
	}
	else if (new->WorkY + new->WorkH < new->ClipH)
	{
	    new->WorkY = new->ClipH - new->WorkH;
	}
	switch(SW_Placement(w))
	{
	case XmTOP_LEFT:
	    if (new->ShowHSB && new->ShowVSB)
	    {
		new->HsbX = SW_XOffset(w) + new->VsbW + SW_Spacing(w);
		new->HsbY = SW_YOffset(w);
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w) + new->HsbH + SW_Spacing(w);
		new->ClipX = new->HsbX + hsb_ht + sw_st;
		new->ClipY = new->VsbY + vsb_ht + sw_st;
	    }
	    else if (new->ShowHSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w);
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->ClipX = new->HsbX + hsb_ht + sw_st;
		new->ClipY = SW_YOffset(w) + new->HsbH + SW_Spacing(w) + vsb_ht + sw_st;
	    }
	    else if (new->ShowVSB)
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w);
		new->ClipX = SW_XOffset(w) + new->VsbW + SW_Spacing(w) + hsb_ht +sw_st;
		new->ClipY = new->VsbY + vsb_ht + sw_st;
	    }
	    else
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->ClipX = SW_XOffset(w) + sw_st;
		new->ClipY = SW_YOffset(w) + sw_st;
	    }
	    break;
	case XmBOTTOM_LEFT:
	    if (new->ShowHSB && new->ShowVSB)
	    {
		new->HsbX = SW_XOffset(w) + new->VsbW + SW_Spacing(w);
		new->HsbY = SW_YOffset(w) + new->VsbH + SW_Spacing(w);
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w);
		new->ClipX = new->HsbX + hsb_ht + sw_st;
		new->ClipY = new->VsbY + vsb_ht + sw_st;
	    }
	    else if (new->ShowHSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w) + usable_height - new->HsbH;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->ClipX = new->HsbX + hsb_ht + sw_st;
		new->ClipY = SW_YOffset(w) + sw_st;
	    }
	    else if (new->ShowVSB)
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = SW_XOffset(w);
		new->VsbY = SW_YOffset(w);
		new->ClipX = SW_XOffset(w) + sw_st;
		new->ClipY = new->VsbY + vsb_ht + sw_st;
	    }
	    else
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->ClipX = SW_XOffset(w) + sw_st;
		new->ClipY = SW_YOffset(w) + sw_st;
	    }
	    break;
	case XmTOP_RIGHT:
	    if (new->ShowHSB && new->ShowVSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w);
		new->VsbX = SW_XOffset(w) + usable_width - new->VsbW;
		new->VsbY = SW_YOffset(w) + new->HsbH + SW_Spacing(w);
		new->ClipX = new->HsbX + hsb_ht + sw_st;
		new->ClipY = new->VsbY + vsb_ht + sw_st;
	    }
	    else if (new->ShowHSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w);
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->ClipX = new->HsbX + hsb_ht + sw_st;
		new->ClipY = SW_YOffset(w) + new->HsbH + SW_Spacing(w) + vsb_ht + sw_st;
	    }
	    else if (new->ShowVSB)
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = SW_XOffset(w) + usable_width - new->VsbW;
		new->VsbY = SW_YOffset(w);
		new->ClipX = SW_XOffset(w) + sw_st;
		new->ClipY = new->VsbY + vsb_ht + sw_st;
	    }
	    else
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->ClipX = SW_XOffset(w) + sw_st;
		new->ClipY = SW_YOffset(w) + sw_st;
	    }
	    break;
	case XmBOTTOM_RIGHT:
	    if (new->ShowHSB && new->ShowVSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w) + usable_height - new->HsbH;
		new->VsbX = SW_XOffset(w) + usable_width - new->VsbW;
		new->VsbY = SW_YOffset(w);
		new->ClipX = new->HsbX + hsb_ht + sw_st;
		new->ClipY = new->VsbY + vsb_ht + sw_st;
	    }
	    else if (new->ShowHSB)
	    {
		new->HsbX = SW_XOffset(w);
		new->HsbY = SW_YOffset(w) + usable_height - new->HsbH;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->ClipX = new->HsbX + hsb_ht + sw_st;
		new->ClipY = SW_YOffset(w) + sw_st;
	    }
	    else if (new->ShowVSB)
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = SW_XOffset(w) + usable_width - new->VsbW;
		new->VsbY = SW_YOffset(w);
		new->ClipX = SW_XOffset(w) + sw_st;
		new->ClipY = new->VsbY + vsb_ht + sw_st;
	    }
	    else
	    {
		new->HsbX = new->SwW;
		new->HsbY = 0;
		new->VsbX = new->SwW;
		new->VsbY = 0;
		new->ClipX = SW_XOffset(w) + sw_st;
		new->ClipY = SW_YOffset(w) + sw_st;
	    }
	    break;
	default:
	}
    }
}

static void
Layout(Widget w, Widget child, XtWidgetGeometry *wants, XmSWValues *new)
{
    XmSWValues current;
    Boolean has_work, has_clip;
    Dimension wk_bw;

    has_clip = SW_ClipWindow(w) != NULL;
    has_work = SW_WorkWindow(w) != NULL;
    wk_bw = has_work ? 2 * XtBorderWidth(SW_WorkWindow(w)) : 0;
    current.HasHSB = SW_HSB(w) != NULL;
    current.ShowHSB = current.HasHSB  && XtIsManaged(SW_HSB(w));
    current.HasVSB = SW_VSB(w) != NULL;
    current.ShowVSB = current.HasVSB && XtIsManaged(SW_VSB(w));
    current.VsbW = current.HasVSB ? XtWidth(SW_VSB(w)) : 0;
    current.VsbH = current.HasVSB ? XtHeight(SW_VSB(w)) : 0;
    current.HsbW = current.HasHSB ? XtWidth(SW_HSB(w)) : 0;
    current.HsbH = current.HasHSB ? XtHeight(SW_HSB(w)) : 0;
    current.ClipW = has_clip ? XtWidth(SW_ClipWindow(w)) : 0;
    current.ClipH = has_clip ? XtHeight(SW_ClipWindow(w)) : 0;
    current.ClipX = has_clip ? XtX(SW_ClipWindow(w)) : 0;
    current.ClipY = has_clip ? XtY(SW_ClipWindow(w)) : 0;
    current.WorkW = has_work ? XtWidth(SW_WorkWindow(w)) + wk_bw : 0;
    current.WorkH = has_work ? XtHeight(SW_WorkWindow(w)) + wk_bw : 0;
    current.WorkX = has_work ? XtX(SW_WorkWindow(w)) : 0;
    current.WorkY = has_work ? XtY(SW_WorkWindow(w)) : 0;
    current.SwW = XtWidth(w);
    current.SwH = XtHeight(w);
    if (child != NULL && wants != NULL)
    {
	if (child == (Widget) SW_ClipWindow(w))
	{
	    current.ClipW = wants->width;
	    current.ClipH = wants->height;
	    current.ClipX = wants->x;
	    current.ClipY = wants->y;
	}
	else if (child == (Widget) SW_WorkWindow(w))
	{
	    current.WorkW = wants->width + 2 * wants->border_width;
	    current.WorkH = wants->height + 2 * wants->border_width;
	    current.WorkX = wants->x;
	    current.WorkY = wants->y;
	}
	else if (child == (Widget) SW_HSB(w))
	{
	    current.HsbW = wants->width;
	    current.HsbH = wants->height;
	}
	else if (child == (Widget) SW_VSB(w))
	{
	    current.VsbW = wants->width;
	    current.VsbH = wants->height;
	}
    }
    if (SW_VisualPolicy(w) == XmCONSTANT)
    {
	ConstantLayout(w, &current, new);
    }
    else
    {
	VariableLayout(w, &current, new);
    }
}


static void
ConfigureChildren(Widget w, Widget child, XmSWValues *vals)
{

    if (XdbInDebug(__FILE__, w))
    {
	DEBUGOUT(XdbDebug(__FILE__, w,
			  "ConfigureChildren: Sw %dx%d\n",
			  vals->SwW, vals->SwH));
	DEBUGOUT(XdbDebug0(__FILE__, w,
			   "\t\tX\tY\tW\tH\n\tWork\t%d\t%d\t%d\t%d\n",
			   vals->WorkX, vals->WorkY, vals->WorkW, vals->WorkH));
	DEBUGOUT(XdbDebug0(__FILE__, w,
			   "\tClip\t%d\t%d\t%d\t%d\n",
			   vals->ClipX, vals->ClipY, vals->ClipW, vals->ClipH));
	DEBUGOUT(XdbDebug0(__FILE__, w,
			   "\tHsb\t%d\t%d\t%d\t%d\t%3s managed\n",
			   vals->HsbX, vals->HsbY, vals->HsbW, vals->HsbH,
			   (vals->ShowHSB && vals->HasHSB) ? "" : "not"));
	DEBUGOUT(XdbDebug0(__FILE__, w,
			   "\tVsb\t%d\t%d\t%d\t%d\t%3s managed\n",
			   vals->VsbX, vals->VsbY, vals->VsbW, vals->VsbH,
			   (vals->ShowVSB && vals->HasVSB) ? "" : "not"));
    }

    /* Clear the old shadow */
    if (XtIsRealized(w))
    {
	if (SW_VisualPolicy(w) == XmCONSTANT)
	{
	    Dimension st = MGR_ShadowThickness(w);
	    Dimension bw = XtBorderWidth(SW_ClipWindow(w));


	    _XmClearBorder(XtDisplay(w), XtWindow(w),
			   XtX(SW_ClipWindow(w)) - st,
			   XtY(SW_ClipWindow(w)) - st,
			   XtWidth(SW_ClipWindow(w)) + 2 * (bw + st),
			   XtHeight(SW_ClipWindow(w)) + 2 * (bw + st),
			   st);
	}
	else if (SW_WorkWindow(w) != NULL && XtIsManaged(SW_WorkWindow(w)))
	{
	    Dimension st = MGR_ShadowThickness(w);
	    Dimension bw = XtBorderWidth(SW_WorkWindow(w));

	    _XmClearBorder(XtDisplay(w), XtWindow(w),
			   XtX(SW_WorkWindow(w)) - st,
			   XtY(SW_WorkWindow(w)) - st,
			   XtWidth(SW_WorkWindow(w)) + 2 * (bw + st),
			   XtHeight(SW_WorkWindow(w)) + 2 * (bw + st),
			   st);
	}
    }
    if (SW_VisualPolicy(w) == XmCONSTANT)
    {

	if (child == (Widget)SW_ClipWindow(w))
	{
	    XtX(SW_ClipWindow(w)) = vals->ClipX;
	    XtY(SW_ClipWindow(w)) = vals->ClipY;
	    XtWidth(SW_ClipWindow(w)) = vals->ClipW;
	    XtHeight(SW_ClipWindow(w)) = vals->ClipH;
	}
	else
	{
	    _XmConfigureObject((Widget)SW_ClipWindow(w),
			       vals->ClipX, vals->ClipY,
			       vals->ClipW, vals->ClipH,
			       0);
	}
	if (SW_WorkWindow(w))
	{
	    Dimension bw = XtBorderWidth(SW_WorkWindow(w));

	    if (child == SW_WorkWindow(w))
	    {
		XtX(SW_WorkWindow(w)) = vals->WorkX;
		XtY(SW_WorkWindow(w)) = vals->WorkY;
		XtWidth(SW_WorkWindow(w)) = vals->WorkW - 2 *bw;
		XtHeight(SW_WorkWindow(w)) = vals->WorkH - 2 *bw;
	    }
	    else
	    {
		_XmConfigureObject(SW_WorkWindow(w),
				   vals->WorkX, vals->WorkY,
				   vals->WorkW - 2 * bw, vals->WorkH - 2 * bw,
				   bw);
	    }
	}
	
	SW_CWWidth(w) = vals->ClipW;
	SW_CWHeight(w) = vals->ClipH;
    }
    else
    {
	if (SW_WorkWindow(w))
	{
	    Dimension bw = XtBorderWidth(SW_WorkWindow(w));

	    if (child == SW_WorkWindow(w))
	    {
		XtX(SW_WorkWindow(w)) = vals->WorkX;
		XtY(SW_WorkWindow(w)) = vals->WorkY;
		XtWidth(SW_WorkWindow(w)) = vals->WorkW - 2 * bw;
		XtHeight(SW_WorkWindow(w)) = vals->WorkH - 2 * bw;
	    }
	    else
	    {
		_XmConfigureObject(SW_WorkWindow(w),
				   vals->WorkX, vals->WorkY,
				   vals->WorkW - 2 * bw, vals->WorkH - 2 * bw,
				   bw);
	    }
	}
	SW_CWWidth(w) = vals->WorkW;
	SW_CWHeight(w) = vals->WorkH;
    }
    if (SW_HSB(w))
    {
	if (child == (Widget) SW_HSB(w))
	{
	    XtX(SW_HSB(w)) = vals->HsbX;
	    XtY(SW_HSB(w)) = vals->HsbY;
	    XtWidth(SW_HSB(w)) = vals->HsbW;
	    XtHeight(SW_HSB(w)) = vals->HsbH;
	}
	else
	{
	    _XmConfigureObject((Widget) SW_HSB(w),
			       vals->HsbX, vals->HsbY,
			       vals->HsbW, vals->HsbH,
			       0);
	}
	if (!vals->ShowHSB && _XmFocusIsHere((Widget) SW_HSB(w)))
	{
	    XmProcessTraversal(w, XmTRAVERSE_NEXT_TAB_GROUP);
	}
    }
    SW_HSBX(w) = vals->HsbX;
    SW_HSBY(w) = vals->HsbY;
    SW_HSBWidth(w) = vals->HsbW;
    SW_HSBHeight(w) = vals->HsbH;
    SW_HasHSB(w) = vals->ShowHSB;
    if (SW_VSB(w))
    {
	if (child == (Widget) SW_VSB(w))
	{
	    XtX(SW_VSB(w)) = vals->VsbX;
	    XtY(SW_VSB(w)) = vals->VsbY;
	    XtWidth(SW_VSB(w)) = vals->VsbW;
	    XtHeight(SW_VSB(w)) = vals->VsbH;
	}
	else
	{
	    _XmConfigureObject((Widget) SW_VSB(w),
			       vals->VsbX, vals->VsbY,
			       vals->VsbW, vals->VsbH,
			       0);
	}
	if (!vals->ShowVSB && _XmFocusIsHere((Widget) SW_VSB(w)))
	{
	    XmProcessTraversal(w, XmTRAVERSE_NEXT_TAB_GROUP);
	}
    }
    SW_VSBX(w) = vals->VsbX;
    SW_VSBY(w) = vals->VsbY;
    SW_VSBWidth(w) = vals->VsbW;
    SW_VSBHeight(w) = vals->VsbH;
    SW_HasVSB(w) = vals->ShowVSB;

    if (SW_VisualPolicy(w) == XmCONSTANT)
    {
	_XmInitializeScrollBars(w);
    }
}

static void
resize(Widget w)
{
    XmSWValues vals;

    DEBUGOUT(XdbDebug(__FILE__, w, "Resize: x %d y %d w %d h %d\n",
		      XtX(w), XtY(w), XtWidth(w), XtHeight(w)));

    SW_FromResize(w) = True;

    /* resize not chained */
    Layout(w, NULL, NULL, &vals);
    ConfigureChildren(w, NULL, &vals);
    SW_FromResize(w) = False;
}

static void
expose(Widget w, XEvent *event, Region region)
{

    _XmRedisplayGadgets(w, event, region);

    if (SW_ScrollPolicy(w) == XmAUTOMATIC)
    {
	_XmDrawShadows(XtDisplay(w), XtWindow(w),
		       MGR_TopShadowGC(w), MGR_BottomShadowGC(w),
		       XtX(SW_ClipWindow(w)) - MGR_ShadowThickness(w),
		       XtY(SW_ClipWindow(w)) - MGR_ShadowThickness(w),
		       SW_CWWidth(w) + 2 * MGR_ShadowThickness(w),
		       SW_CWHeight(w) + 2 * MGR_ShadowThickness(w),
		       MGR_ShadowThickness(w), XmSHADOW_IN);
    }
    else
    {
	_XmDrawShadows(XtDisplay(w), XtWindow(w),
		       MGR_TopShadowGC(w), MGR_BottomShadowGC(w),
		       XtX(SW_WorkWindow(w)) - MGR_ShadowThickness(w),
		       XtY(SW_WorkWindow(w)) - MGR_ShadowThickness(w),
		       SW_CWWidth(w) + 2 * MGR_ShadowThickness(w),
		       SW_CWHeight(w) + 2 * MGR_ShadowThickness(w),
		       MGR_ShadowThickness(w), XmSHADOW_IN);
    }
}


/*
 * Allow other widgets to ask what the preferred geometry of this widget is.
 */
static XtGeometryResult
query_geometry(Widget w,
	       XtWidgetGeometry *intended,
	       XtWidgetGeometry *preferred)
{
    XtWidgetGeometry request;

    DEBUGOUT(XdbDebug(__FILE__, w, "query_geometry(%s)\n",
		      XdbWidgetGeometry2String(intended)));
    if (SW_WorkWindow(w) == NULL || !XtIsManaged(SW_WorkWindow(w)))
    {
	preferred->height = XtHeight(w);
	preferred->width = XtWidth(w);
	if ((intended->request_mode & (CWWidth | CWHeight)) ==
	    (CWWidth | CWHeight))
	{
	    preferred->request_mode = 0;
	}
	else
	{
	    preferred->request_mode = CWWidth | CWHeight;

	    if ((intended->request_mode & (CWWidth | CWHeight)) == CWWidth)
	    {
		preferred->request_mode &= ~CWWidth;
	    }
	    if ((intended->request_mode & (CWWidth | CWHeight)) == CWHeight)
	    {
		preferred->request_mode &= ~CWHeight;
	    }
	}
    }
    else
    {
	request.request_mode =
	    intended->request_mode & (CWWidth | CWHeight);
	if (!XtIsRealized(w) && 
	    ((intended->request_mode & (CWWidth | CWHeight)) != 0))
	{
	    preferred->width = XtWidth(w);
	    preferred->height = XtHeight(w);
	}
	else
	{
	    request.width = intended->width;
	    request.height = intended->height;
	    _XmScrolledWindowPreferredSize(w, &request, preferred);
	    if ((intended->request_mode & (CWWidth | CWHeight)) == CWWidth)
	    {
		preferred->request_mode &= ~CWWidth;
	    }
	    if ((intended->request_mode & (CWWidth | CWHeight)) == CWHeight)
	    {
		preferred->request_mode &= ~CWHeight;
	    }
	}
    }
    return _XmGMReplyToQueryGeometry(w, intended, preferred);
}


/* The rules of the games is this:
 * Horizontal ScrollBar can change height, provided SW can change its
 * height to fit.
 * Vertical ScrollBar can change width, provided SW can change its 
 * width to fit.
 * WorkWindow can change x, y, width, height and border_width if visualPolicy
 * is XmCONSTANT, and it dosnt break the ScrollBar constraints.
 * If visualPolicy is XmVARIABLE it can change width, height and border_width,
 * provided SW can change its size to fit.
 * If visualPolicy is Constant ClipWindow substitute for the workwindow.
 * Any other child cant change anything.
 */

static XtGeometryResult
geometry_manager(Widget w, XtWidgetGeometry *desired, XtWidgetGeometry *allowed)
{
    XtWidgetGeometry wants, request, answer;
    Widget sw = XtParent(w);
    XmSWValues vals;
    XtGeometryResult result = XtGeometryYes;
    XtGeometryMask query_only;

    DEBUGOUT(XdbDebug2(__FILE__, sw, w, "geometry_manager (request %s)\n",
		       XdbWidgetGeometry2String(desired)));

#define	Wants(flag)	(desired->request_mode & flag)


    wants = *desired;
    query_only = wants.request_mode & XtCWQueryOnly;
    allowed->request_mode = query_only;


    allowed->x = (wants.request_mode & CWX) ? wants.x: XtX(w);
    allowed->y = (wants.request_mode & CWY) ? wants.y: XtY(w);
    allowed->width = (wants.request_mode & CWWidth) ?
	wants.width: XtWidth(w);
    allowed->height = (wants.request_mode & CWHeight) ?
	wants.height: XtHeight(w);
    allowed->border_width = (wants.request_mode & CWBorderWidth) ?
	wants.border_width: XtBorderWidth(w);

    if (w != (Widget) SW_WorkWindow(sw) &&
	w != (Widget) SW_ClipWindow(sw) &&
	w != (Widget) SW_HSB(sw) &&
	w != (Widget) SW_VSB(sw))
    {
	if (!query_only)
	{
	    XtX(w) = allowed->x;
	    XtY(w) = allowed->y;
	    XtWidth(w) = allowed->width;
	    XtHeight(w) = allowed->height;
	    XtBorderWidth(w) = allowed->border_width;
	}
	return XtGeometryYes;
    }



    if (w == (Widget)SW_ClipWindow(sw) && SW_VisualPolicy(sw) == XmCONSTANT)
    {
	Dimension clip_width = XtWidth(SW_ClipWindow(sw));
	Dimension clip_height = XtHeight(SW_ClipWindow(sw));
	
	DEBUGOUT(XdbDebug2(__FILE__, sw, w,
			   "BEGIN AUTOMATIC FAKE\n\n"));
	
	DEBUGOUT(XdbDebug2(__FILE__, sw, w,
			   "geometry_manager: resize WorkWindow: %d %d"
			   " %d %d %d\n",
			   wants.x, wants.y,
			   wants.width, wants.height, wants.border_width));

	if (SW_WorkWindow(sw) != NULL)
	{
	    w = SW_WorkWindow(sw);
	    allowed->x = (wants.request_mode & CWX) ? wants.x: XtX(w);
	    allowed->y = (wants.request_mode & CWY) ? wants.y: XtY(w);
	    allowed->width = (wants.request_mode & CWWidth) ?
		wants.width: XtWidth(w);
	    allowed->height = (wants.request_mode & CWHeight) ?
		wants.height: XtHeight(w);
	    allowed->border_width = (wants.request_mode & CWBorderWidth) ?
		wants.border_width: XtBorderWidth(w);
	}
	else
	{
	    return XtGeometryNo;
	}

	if (allowed->width <= clip_width)
	{
	    allowed->x = 0;
	}
	else if (allowed->x + allowed->width + 2 * allowed->border_width <
		 clip_width)
	{
	    allowed->x = clip_width - allowed->width - 
		2 * allowed->border_width;
	}
	if (allowed->height + 2 * allowed->border_width <= clip_height)
	{
	    allowed->y = 0;
	}
	else if (allowed->y + allowed->height + 2 * allowed->border_width 
		 < clip_height)
	{
	    allowed->y = clip_height - allowed->height -
		2 * allowed->border_width;
	}
	if ((wants.request_mode & CWX) && allowed->x != wants.x)
	{
	    result = XtGeometryAlmost;
	    allowed->request_mode |= CWX;
	}
	if ((wants.request_mode & CWY) && allowed->y != wants.y)
	{
	    result = XtGeometryAlmost;
	    allowed->request_mode |= CWY;
	}
    }
    if (w == (Widget)SW_WorkWindow(sw) && SW_VisualPolicy(sw) == XmVARIABLE)
    {
	if ((wants.request_mode & ~(CWX | CWY | XtCWQueryOnly)) == 0)
	{
	    result = XtGeometryNo;
	}
	else
	{
	    request.request_mode = CWWidth | CWHeight | query_only;
	    request.width = XtWidth(sw);
	    request.height = XtHeight(sw);
	    if (wants.request_mode & CWWidth)
	    {
		request.width += wants.width - XtWidth(w);
	    }
	    if (wants.request_mode & CWHeight)
	    {
		request.height += wants.height - XtHeight(w);
	    }
	    if (wants.request_mode & CWBorderWidth)
	    {
		request.width += 
		    2 * (wants.border_width - XtBorderWidth(w));
		request.height += 
		    2 * (wants.border_width - XtBorderWidth(w));
	    }
	    result = XtMakeGeometryRequest(sw, &request, &answer);
	    if (result == XtGeometryAlmost)
	    {
		if ((wants.request_mode & CWWidth) &&
		    request.width != answer.width)
		{
		    allowed->width = wants.width + answer.width -
			request.width;
		    allowed->request_mode |= CWWidth;
		    if (wants.request_mode & CWBorderWidth)
		    {
			result = XtGeometryNo;
		    }
		}
		else if ((wants.request_mode & CWHeight) &&
			 request.height != answer.width)
		{
		    allowed->height = wants.height + answer.height -
			request.height;
		    allowed->request_mode |= CWHeight;
		    if (wants.request_mode & CWBorderWidth)
		    {
			result = XtGeometryNo;
		    }
		} 
		else if (wants.request_mode & CWBorderWidth)
		{
		    result = XtGeometryNo;
		}
	    }
	}
    }
    else if (w == (Widget) SW_HSB(sw))
    {
	if ((wants.request_mode & 
	     ~(CWX | CWY | CWWidth | CWBorderWidth | XtCWQueryOnly)) == 0)
	{
	    result = XtGeometryNo;
	}
	else
	{
	    request.request_mode = CWHeight | query_only;
	    request.height = XtHeight(sw) + wants.height - XtHeight(w);
	    result = XtMakeGeometryRequest(sw, &request, &answer);
	    if (result == XtGeometryAlmost)
	    {
		allowed->request_mode |= CWHeight; 	    
		allowed->height = wants.height + answer.height -
		    request.height;
	    }
	}
    }
    else if (w == (Widget) SW_VSB(sw))
    {
	if ((wants.request_mode & 
	     ~(CWX | CWY | CWHeight | CWBorderWidth | XtCWQueryOnly)) == 0)
	{
	    result = XtGeometryNo;
	}
	else
	{
	    request.request_mode = CWWidth | query_only;
	    request.width = XtWidth(sw) + wants.width - XtWidth(w);
	    result = XtMakeGeometryRequest(sw, &request, &answer);
	    if (result == XtGeometryAlmost)
	    {
		allowed->request_mode |= CWWidth; 	    
		allowed->width = wants.width + answer.width -
		    request.width;
	    }
	}
    }
    if (result == XtGeometryYes && !query_only)
    {
	Layout(sw, w, allowed, &vals);
	ConfigureChildren(sw, w, &vals);
	if (XtIsRealized(w))
	{
	    if (SW_ScrollPolicy(sw) == XmAUTOMATIC)
	    {
		_XmDrawShadows(XtDisplay(sw), XtWindow(sw),
			       MGR_TopShadowGC(sw), MGR_BottomShadowGC(sw),
			       XtX(SW_ClipWindow(sw)) -
			       MGR_ShadowThickness(sw),
			       XtY(SW_ClipWindow(sw)) -
			       MGR_ShadowThickness(sw),
			       SW_CWWidth(sw) + 2 * MGR_ShadowThickness(sw),
			       SW_CWHeight(sw) + 2 * MGR_ShadowThickness(sw),
			       MGR_ShadowThickness(sw), XmSHADOW_IN);
	    }
	    else
	    {
		if (SW_WorkWindow(sw) != NULL &&
		    XtIsManaged(SW_WorkWindow(sw)))
		{
		    _XmDrawShadows(XtDisplay(sw), XtWindow(sw),
				   MGR_TopShadowGC(sw), MGR_BottomShadowGC(sw),
				   XtX(SW_WorkWindow(sw)) -
				   MGR_ShadowThickness(sw),
				   XtY(SW_WorkWindow(sw)) -
				   MGR_ShadowThickness(sw),
				   SW_CWWidth(sw) +
				   2 * MGR_ShadowThickness(sw),
				   SW_CWHeight(sw) +
				   2 * MGR_ShadowThickness(sw),
				   MGR_ShadowThickness(sw), XmSHADOW_IN);
		}
	    }
	}
    }
    return result;
}

static void
change_managed(Widget w)
{
    XmSWValues vals;

    DEBUGOUT(XdbDebug(__FILE__, w, "ChangeManaged: x %d y %d w %d h %d\n",
		      XtX(w), XtY(w), XtWidth(w), XtHeight(w)));

    SW_FromResize(w) = True;

    /* resize not chained */
    if (SW_VisualPolicy(w) == XmVARIABLE)
    {
	MakeResizeRequest(w);
    }
    Layout(w, NULL, NULL, &vals);
    ConfigureChildren(w, NULL, &vals);

    if (XtIsRealized(w))
    {
	if (SW_ScrollPolicy(w) == XmAUTOMATIC)
	{
	    _XmDrawShadows(XtDisplay(w), XtWindow(w),
			   MGR_TopShadowGC(w), MGR_BottomShadowGC(w),
			   XtX(SW_ClipWindow(w)) - MGR_ShadowThickness(w),
			   XtY(SW_ClipWindow(w)) - MGR_ShadowThickness(w),
			   SW_CWWidth(w) + 2 * MGR_ShadowThickness(w),
			   SW_CWHeight(w) + 2 * MGR_ShadowThickness(w),
			   MGR_ShadowThickness(w), XmSHADOW_IN);
	}
	else if (SW_WorkWindow(w) != NULL && XtIsManaged(SW_WorkWindow(w)))
	{
	    _XmDrawShadows(XtDisplay(w), XtWindow(w),
			   MGR_TopShadowGC(w), MGR_BottomShadowGC(w),
			   XtX(SW_WorkWindow(w)) - MGR_ShadowThickness(w),
			   XtY(SW_WorkWindow(w)) - MGR_ShadowThickness(w),
			   SW_CWWidth(w) + 2 * MGR_ShadowThickness(w),
			   SW_CWHeight(w) + 2 * MGR_ShadowThickness(w),
			   MGR_ShadowThickness(w), XmSHADOW_IN);
	}
    }

    SW_FromResize(w) = False;
}

/*
 * In here we probably need some trickery to make things work.
 * One sensible thing to do is to make the work widget the
 * child of the clip widget. This is ugly, though. That's probably the
 * reason why the X Consortium never wanted to add XtReparentWidget().
 *
 * The advantage of reparenting the work widget is, all you have to do
 * to scroll is tell the work widget where it is. Through the magic of
 * the X window system having window hierarchies, things should work.
 */
static void
insert_child(Widget w)
{
    XmScrolledWindowWidget sw = (XmScrolledWindowWidget)XtParent(w);

#define superclass (&xmManagerClassRec)

    /* Special first case : these are not the work window
     * Note this is especially here for our subclass XmMainWindow */
    if (XmIsSeparator(w) || XmIsSeparatorGadget(w) ||
	(XmIsRowColumn(w) && RC_Type(w) == XmMENU_BAR))
    {
	(*superclass->composite_class.insert_child) (w);
    }
    else if (XmIsScrolledWindow(w))
    {
	/*
	 * Second special case (for XmMainWindow).
	 * Avoid reparenting stuff which is already a scrolledwindow (child).
	 * It's unlikely to become the workarea anyway.
	 * Example : Danny's (Motif) clone of Ghostview.
	 * 28/8/1997.
	 */
	(*superclass->composite_class.insert_child) (w);
    }
    else if (XmIsScrollBar(w))
    {
	(*superclass->composite_class.insert_child) (w);

	switch (SCB_Orientation(w))
	{
	case XmHORIZONTAL:
	    SW_HSB(sw) = (XmScrollBarWidget)w;
	    if (XtIsManaged(w))
	    {
		SW_HasHSB(w) = True;
	    }
	    else
	    {
		SW_HasHSB(w) = False;
	    }
	    break;

	case XmVERTICAL:
	    SW_VSB(sw) = (XmScrollBarWidget)w;
	    if (XtIsManaged(w))
	    {
		SW_HasVSB(w) = True;
	    }
	    else
	    {
		SW_HasVSB(w) = False;
	    }
	    break;

	default:
	    _XmWarning((Widget)sw,
		       "Can't determine ScrollBar orientation in "
		       "ScrolledWindow %s: Not adding",
		       XtName(sw));
	    break;
	}
    }
    else if (XtIsShell(w))
    {
	/* Shells cannot be the work area - so make sure we catch them here */
	(*superclass->composite_class.insert_child) (w);
    }
    /*
     * only we can add a clip window, and we know it's id
     */
    else if (SW_VisualPolicy(sw) == XmCONSTANT && SW_ClipWindow(sw) == NULL)
    {
	(*superclass->composite_class.insert_child) (w);
    }
    /*
     * THIS IS SOO UGLY
     *
     * if we're not XmCONSTANT, don't reparent
     *
     */
    else if (SW_VisualPolicy(sw) == XmCONSTANT)
    {
	DEBUGOUT(XdbDebug2(__FILE__, w, (Widget)sw, "Reparented to %s\n",
			   XtName(SW_ClipWindow(sw))));

	XtParent(w) = (Widget)SW_ClipWindow(sw);

	/*
	 * Now we're reparented. Go on inserting the child as if
	 * nothing happened
	 */
	(*superclass->composite_class.insert_child) (w);

	SW_WorkWindow(sw) = w;
    }
    else if (SW_WorkWindow(sw) == NULL)
    {
	DEBUGOUT(XdbDebug2(__FILE__, (Widget)sw, w, "Child is Work Window\n"));

	(*superclass->composite_class.insert_child) (w);

	SW_WorkWindow(sw) = w;
    }
    else
    {
	DEBUGOUT(XdbDebug2(__FILE__, (Widget)sw, w, "Child is goofy\n"));

	(*superclass->composite_class.insert_child) (w);
    }
#undef superclass
}

static void
delete_child(Widget w)
{
    /* Motif inherits this method */
    Widget sw = XtParent(w);

    if ((Widget)SW_ClipWindow(sw) == w)
    {
	_XmError(sw, "ClipWindow killed");
    }
    else if ((Widget)SW_VSB(sw) == w)
    {
	if (SW_VisualPolicy(sw) == XmCONSTANT)
	{
	    _XmError(sw, "Vertical ScrollBar killed when visual policy is CONSTANT");
	}
	else {
	    SW_VSB(w) = NULL;
	}
    }
    else if ((Widget)SW_HSB(sw) == w)
    {
	if (SW_VisualPolicy(sw) == XmCONSTANT)
	{
	    _XmError(sw, "Horizontal ScrollBar killed when visual policy is CONSTANT");
	}
	else
	{
	    SW_HSB(w) = NULL;
	}
    }

#define superclass (&xmManagerClassRec)
    (*superclass->composite_class.delete_child) (w);
#undef superclass
}

static void
HsbCallback(Widget w, XtPointer client, XtPointer call)
{
    XmScrollBarCallbackStruct *cb = (XmScrollBarCallbackStruct *) call;
    Widget sw = XtParent(w);

    SW_HSBValue(sw) = cb->value;

    DEBUGOUT(XdbDebug2(__FILE__, sw, w, "HsbCallback Hor %d ", cb->value));

    /*
     * Position the work window
     */
    if (SW_WorkWindow(sw))
    {
	XtMoveWidget(SW_WorkWindow(sw), - cb->value, 0);
    }
}
static void
VsbCallback(Widget w, XtPointer client, XtPointer call)
{
    XmScrollBarCallbackStruct *cb = (XmScrollBarCallbackStruct *) call;
    Widget sw = XtParent(w);

    SW_VSBValue(sw) = cb->value;

    DEBUGOUT(XdbDebug2(__FILE__, sw, w, "VsbCallback Vert %d\n", cb->value));

    /*
     * Position the work window
     */
    if (SW_WorkWindow(sw))
    {
	XtMoveWidget(SW_WorkWindow(sw), 0, - cb->value);
    }
}



void
_XmInitializeScrollBars(Widget w)
{
    int Hsb_Increment, Hsb_PageIncrement, Vsb_Increment, Vsb_PageIncrement;
    Dimension wk_bw;
    Boolean has_work;
    
    has_work = 
    wk_bw = has_work ? 2 * XtBorderWidth(SW_WorkWindow(w)) : 0;

    if (SW_WorkWindow(w) == NULL || !XtIsManaged(SW_WorkWindow(w)))
    {
	SW_HSBMinimum(w) = 0;
	SW_HSBMaximum(w) = 100;
	SW_HSBValue(w) = 0;
	SW_HSBSliderSize(w) = 100;
	Hsb_PageIncrement = 90;
	Hsb_Increment = 10;
	SW_VSBMinimum(w) = 0;
	SW_VSBMaximum(w) = 100;
	SW_VSBValue(w) = 0;
	SW_VSBSliderSize(w) = 100;
	Vsb_PageIncrement = 90;
	Vsb_Increment = 10;
    }
    else 
    {
	Dimension wk_width, wk_bw, wk_height;

	wk_bw = 2 * XtBorderWidth(SW_WorkWindow(w));
	wk_width = XtWidth(SW_WorkWindow(w)) + wk_bw;
	wk_height = XtHeight(SW_WorkWindow(w)) + wk_bw;

	if (wk_width < XtWidth(SW_ClipWindow(w)))
	{
	    SW_HSBMinimum(w) = 0;
	    SW_HSBMaximum(w) = wk_width;
	    SW_HSBValue(w) = 0;
	    SW_HSBSliderSize(w) = wk_width;
	    Hsb_Increment = wk_width / 10;
	    if (Hsb_Increment < 1)
	    {
		Hsb_Increment = 1;
	    }
	    Hsb_PageIncrement = wk_width - wk_width / 10;
	    if (Hsb_PageIncrement < 1)
	    {
		Hsb_PageIncrement = 1;
	    }
	}
	else
	{
	    SW_HSBMinimum(w) = 0;
	    SW_HSBMaximum(w) = wk_width;
	    SW_HSBValue(w) = - XtX(SW_WorkWindow(w));
	    SW_HSBSliderSize(w) = XtWidth(SW_ClipWindow(w));
	    Hsb_Increment = wk_width / 10;
	    if (Hsb_Increment < 1)
	    {
		Hsb_Increment = 1;
	    }
	    Hsb_PageIncrement = XtWidth(SW_ClipWindow(w)) -
		XtWidth(SW_ClipWindow(w)) / 10;;
	    if (Hsb_PageIncrement < 1)
	    {
		Hsb_PageIncrement = 1;
	    }
	}
	
	if (wk_height < XtHeight(SW_ClipWindow(w)))
	{
	    SW_VSBMinimum(w) = 0;
	    SW_VSBMaximum(w) = wk_height;
	    SW_VSBValue(w) = 0;
	    SW_VSBSliderSize(w) = wk_height;
	    Vsb_Increment = wk_height / 10;
	    if (Vsb_Increment < 1)
	    {
		Vsb_Increment = 1;
	    }
	    Vsb_PageIncrement = wk_height - wk_height / 10;
	    if (Vsb_PageIncrement < 1)
	    {
		Vsb_PageIncrement = 1;
	    }
	}
	else
	{
	    SW_VSBMinimum(w) = 0;
	    SW_VSBMaximum(w) = wk_height;
	    SW_VSBValue(w) = - XtY(SW_WorkWindow(w));
	    SW_VSBSliderSize(w) = XtHeight(SW_ClipWindow(w));
	    Vsb_Increment = wk_height / 10;
	    if (Vsb_Increment < 1)
	    {
		Vsb_Increment = 1;
	    }
	    Vsb_PageIncrement = XtHeight(SW_ClipWindow(w)) -
		XtHeight(SW_ClipWindow(w)) / 10;;
	    if (Vsb_PageIncrement < 1)
	    {
		Vsb_PageIncrement = 1;
	    }
	}
    }
    XtVaSetValues((Widget) SW_HSB(w),
		  XmNminimum, SW_HSBMinimum(w),
		  XmNmaximum, SW_HSBMaximum(w),
		  XmNsliderSize, SW_HSBSliderSize(w),
		  XmNvalue, SW_HSBValue(w),
		  XmNincrement, Hsb_Increment,
		  XmNpageIncrement, Hsb_PageIncrement,
		  NULL);
	
    XtVaSetValues((Widget) SW_VSB(w),
		  XmNminimum, SW_VSBMinimum(w),
		  XmNmaximum, SW_VSBMaximum(w),
		  XmNsliderSize, SW_VSBSliderSize(w),
		  XmNvalue, SW_VSBValue(w),
		  XmNincrement, Vsb_Increment,
		  XmNpageIncrement, Vsb_PageIncrement,
		  NULL);
}




static void
PageUp(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageUp\n"));

    if (SW_ScrollPolicy(w) == XmAUTOMATIC && SW_WorkWindow(w) != NULL)
    {
	Dimension ht_c = XtHeight(SW_ClipWindow(w));
	Dimension ht_w = XtHeight(SW_WorkWindow(w)) +
	    2 *  XtBorderWidth(SW_WorkWindow(w));

	if (ht_c < ht_w)
	{
	    Position y = XtY(SW_WorkWindow(w)) + ht_c - ht_c / 10;

	    if (y > 0)
	    {
		y = 0;
	    }
	    XtMoveWidget(SW_WorkWindow(w), XtX(SW_WorkWindow(w)), y);
	    SW_VSBValue(w) = - y;
	    XtVaSetValues((Widget) SW_VSB(w), XmNvalue, - y, NULL);
	}
    }
}

static void
PageDown(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageDown\n"));

    if (SW_ScrollPolicy(w) == XmAUTOMATIC && SW_WorkWindow(w) != NULL)
    {
	Dimension ht_c = XtHeight(SW_ClipWindow(w));
	Dimension ht_w = XtHeight(SW_WorkWindow(w)) + 
	    2 *  XtBorderWidth(SW_WorkWindow(w));

	if (ht_c < ht_w)
	{
	    Position y = XtY(SW_WorkWindow(w)) - ht_c + ht_c / 10;

	    if (y < ht_c - ht_w)
	    {
		y = ht_c - ht_w;
	    }
	    XtMoveWidget(SW_WorkWindow(w), XtX(SW_WorkWindow(w)), y);
	    SW_VSBValue(w) = - y;
	    XtVaSetValues((Widget) SW_VSB(w), XmNvalue, - y, NULL);
	}
    }
}


static void
PageLeft(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageLeft\n"));
    if (SW_ScrollPolicy(w) == XmAUTOMATIC && SW_WorkWindow(w) != NULL)
    {
 	Dimension wt_c = XtWidth(SW_ClipWindow(w));
	Dimension wt_w = XtWidth(SW_WorkWindow(w)) +
	    2 *  XtBorderWidth(SW_WorkWindow(w));

	if (wt_c < wt_w)
	{
	    Position x = XtX(SW_WorkWindow(w)) + wt_c - wt_c / 10;

	    if (x > 0)
	    {
		x = 0;
	    }
	    XtMoveWidget(SW_WorkWindow(w), x, XtY(SW_WorkWindow(w)));
	    SW_HSBValue(w) = - x;
	    XtVaSetValues((Widget) SW_HSB(w), XmNvalue, - x, NULL);
	}
    }
}

static void
PageRight(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageRight\n"));

    if (SW_ScrollPolicy(w) == XmAUTOMATIC && SW_WorkWindow(w) != NULL)
    {
 	Dimension wt_c = XtWidth(SW_ClipWindow(w));
	Dimension wt_w = XtWidth(SW_WorkWindow(w)) + 
	    2 *  XtBorderWidth(SW_WorkWindow(w));

	if (wt_c < wt_w)
	{
	    Position x = XtX(SW_WorkWindow(w)) - wt_c + wt_c / 10;

	    if (x < wt_c - wt_w)
	    {
		x = wt_c - wt_w;
	    }
	    XtMoveWidget(SW_WorkWindow(w), x, XtY(SW_WorkWindow(w)));
	    SW_HSBValue(w) = - x;
	    XtVaSetValues((Widget) SW_HSB(w), XmNvalue, - x, NULL);
	}
    }
}

static void
BeginLine(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "BeginLine\n"));

    if (SW_ScrollPolicy(w) == XmAUTOMATIC && SW_WorkWindow(w) != NULL)
    {
 	Dimension wt_c = XtWidth(SW_ClipWindow(w));
	Dimension wt_w = XtWidth(SW_WorkWindow(w)) + 
	    2 *  XtBorderWidth(SW_WorkWindow(w));

	if (wt_c < wt_w)
	{
	    XtMoveWidget(SW_WorkWindow(w), 0, XtY(SW_WorkWindow(w)));
	    SW_HSBValue(w) = 0;
	    XtVaSetValues((Widget) SW_HSB(w), XmNvalue, 0, NULL);
	}
    }
}

static void
EndLine(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageRight\n"));
    
    if (SW_ScrollPolicy(w) == XmAUTOMATIC && SW_WorkWindow(w) != NULL)
    {
 	Dimension wt_c = XtWidth(SW_ClipWindow(w));
	Dimension wt_w = XtWidth(SW_WorkWindow(w)) + 
	    2 *  XtBorderWidth(SW_WorkWindow(w));

	if (wt_c < wt_w)
	{
	    Position x = wt_c - wt_w;
	    XtMoveWidget(SW_WorkWindow(w), x, XtY(SW_WorkWindow(w)));
	    SW_HSBValue(w) = - x;
	    XtVaSetValues((Widget) SW_HSB(w), XmNvalue, - x, NULL);
	}
    }
}

static void
BeginData(Widget w, XEvent *event, String *params, Cardinal *num_params)
{

    DEBUGOUT(XdbDebug(__FILE__, w, "BeginData\n"));

    if (SW_ScrollPolicy(w) == XmAUTOMATIC && SW_WorkWindow(w) != NULL)
    {
	Dimension ht_c = XtHeight(SW_ClipWindow(w));
	Dimension ht_w = XtHeight(SW_WorkWindow(w)) + 
	    2 *  XtBorderWidth(SW_WorkWindow(w));

	if (ht_c < ht_w)
	{
	    XtMoveWidget(SW_WorkWindow(w), XtX(SW_WorkWindow(w)), 0);
	    SW_VSBValue(w) = 0;
	    XtVaSetValues((Widget) SW_VSB(w), XmNvalue, 0, NULL);
	}
    }
}

static void
EndData(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "EndData\n"));

    if (SW_ScrollPolicy(w) == XmAUTOMATIC && SW_WorkWindow(w) != NULL)
    {
 	Dimension ht_c = XtHeight(SW_ClipWindow(w));
	Dimension ht_w = XtHeight(SW_WorkWindow(w)) + 
	    2 *  XtBorderWidth(SW_WorkWindow(w));

	if (ht_c < ht_w)
	{
	    Position y = ht_c - ht_w;

	    XtMoveWidget(SW_WorkWindow(w), XtX(SW_WorkWindow(w)), y);
	    SW_VSBValue(w) = - y;
	    XtVaSetValues((Widget) SW_VSB(w), XmNvalue, - y, NULL);
	}
    }
}

static void
PageUpGrab(Widget w, XEvent *event,
	   String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageUpGrab\n"));

    PageUp(XtParent(w), event, params, num_params);
    XtUngrabKeyboard(w, event->xkey.time);
}

static void
PageDownGrab(Widget w, XEvent *event,
	     String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageDownGrab\n"));

    PageDown(XtParent(w), event, params, num_params);
    XtUngrabKeyboard(w, event->xkey.time);
}

static void
PageLeftGrab(Widget w, XEvent *event,
	     String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageDown\n"));

    PageLeft(XtParent(w), event, params, num_params);
    XtUngrabKeyboard(w, event->xkey.time);
}

static void
PageRightGrab(Widget w, XEvent *event,
	      String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "PageRightGrab\n"));

    PageRight(XtParent(w), event, params, num_params);
    XtUngrabKeyboard(w, event->xkey.time);
}

static void
BeginLineGrab(Widget w, XEvent *event,
	      String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "BeginLineGrab\n"));

    BeginLine(XtParent(w), event, params, num_params);
    XtUngrabKeyboard(w, event->xkey.time);
}

static void
EndLineGrab(Widget w, XEvent *event,
	    String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "EndLineGrab\n"));

    EndLine(XtParent(w), event, params, num_params);
    XtUngrabKeyboard(w, event->xkey.time);
}

static void
BeginDataGrab(Widget w, XEvent *event,
	      String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "BeginDataGrab\n"));

    BeginData(XtParent(w), event, params, num_params);
    XtUngrabKeyboard(w, event->xkey.time);
}

static void
EndDataGrab(Widget w, XEvent *event,
	    String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "EndDataGrab\n"));

    EndData(XtParent(w), event, params, num_params);
    XtUngrabKeyboard(w, event->xkey.time);
}

static void
SWNoop(Widget w, XEvent *event,
       String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "SWNoop\n"));
}

static void
CWMapNotify(Widget w, XEvent *event,
       String *params, Cardinal *num_params)
{
    DEBUGOUT(XdbDebug(__FILE__, w, "CWMapNotify\n"));

    XtOverrideTranslations(w, TranslationsParsed);
}

Widget
XmCreateScrolledWindow(Widget parent, char *name,
		       Arg *argList, Cardinal argcount)
{
    return XtCreateWidget(name, xmScrolledWindowWidgetClass, parent,
			  argList, argcount);
}

void
XmScrollVisible(Widget scrollw_widget, Widget widget,
		Dimension left_right_margin, Dimension top_bottom_margin)
{
    Position cx, cy, wx, wy;
    int cw, ch, ww, wh;
    int dx, dy;
    int val;
    Widget par;

    /* sanity checks */
    if (SW_ScrollPolicy(scrollw_widget) != XmAUTOMATIC)
    {
	return;
    }

    if (!SW_ClipWindow(scrollw_widget) || !SW_WorkWindow(scrollw_widget))
    {
	return;
    }

    for (par = widget; !XtIsShell(par); par = XtParent(par))
    {
	if (par == scrollw_widget)
	{
	    break;
	}
    }

    if (par != scrollw_widget)
    {
	return;
    }

    /* translate the clip and child coords */
    XtTranslateCoords((Widget)SW_ClipWindow(scrollw_widget), 0, 0, &cx, &cy);
    XtTranslateCoords(widget, 0, 0, &wx, &wy);

    cw = XtWidth(SW_ClipWindow(scrollw_widget));
    ch = XtHeight(SW_ClipWindow(scrollw_widget));
    ww = XtWidth(widget);
    wh = XtHeight(widget);

    /* delta */
    dx = dy = 0;

    /* figure dx */
    if (wx < cx)
    {
	dx = cx - wx + left_right_margin;
    }
    else if (wx + ww > cx + cw)
    {
	dx = -((wx + ww) - (cx + cw) + left_right_margin);
    }
    /* figure dy */
    if (wy < cy)
    {
	dy = cy - wy + top_bottom_margin;
    }
    else if (wy + wh > cy + ch)
    {
	dy = -((wy + wh) - (cy + ch) + top_bottom_margin);
    }

    _XmMoveObject(SW_WorkWindow(scrollw_widget),
		  XtX(SW_WorkWindow(scrollw_widget)) + dx,
		  XtY(SW_WorkWindow(scrollw_widget)) + dy);

    if (SW_HSB(scrollw_widget))
    {
    int min, max;

	XtVaGetValues((Widget)SW_HSB(scrollw_widget), 
		XmNvalue, &val, 
		XmNmaximum, &max, 
		XmNminimum, &min, 
		NULL);
	val -= dx;
	val = val < min ? min : val;
	val = val > max ? max : val;
	XtVaSetValues((Widget)SW_HSB(scrollw_widget), XmNvalue, val, NULL);
    }
    if (SW_VSB(scrollw_widget))
    {
    int min, max;

	XtVaGetValues((Widget)SW_VSB(scrollw_widget), 
		XmNvalue, &val, 
		XmNmaximum, &max, 
		XmNminimum, &min, 
		NULL);
	val -= dy;
	val = val < min ? min : val;
	val = val > max ? max : val;
	XtVaSetValues((Widget)SW_VSB(scrollw_widget), XmNvalue, val, NULL);
    }
}

void
XmScrolledWindowSetAreas(Widget widget,
			 Widget h_scrollbar,
			 Widget v_scrollbar,
			 Widget work_region)
{
    if (h_scrollbar)
    {
	XtVaSetValues(widget,
		  XmNhorizontalScrollBar, h_scrollbar,
		  NULL);
    }
    if (v_scrollbar)
    {
	XtVaSetValues(widget,
		  XmNverticalScrollBar, v_scrollbar,
		  NULL);
    }
    if (work_region)
    {
	XtVaSetValues(widget,
		  XmNworkWindow, work_region,
		  NULL);
    }
}


